/*
 * Total.h
 *
 *  Created on: 20. 9. 2014
 *	  Author: Tomas Pecka
 */

#ifndef TOTAL_H_
#define TOTAL_H_

#include <core/multipleDispatch.hpp>
#include "automaton/Automaton.h"
#include "automaton/AutomatonFeatures.h"

#include <exception/CommonException.h>

#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>
#include <label/FailStateLabel.h>
#include <common/createUnique.hpp>

namespace automaton {

namespace simplify {

/**
 * Makes finite automaton's transition function total.
 * Source: Melichar: Algorithm 2.22
 */
class Total : public alib::SingleDispatch<Total, automaton::Automaton, const automaton::AutomatonBase &> {
public:
	/**
	 * Computes epsilon closure of a state in epsilon nonfree automaton
	 */
	static automaton::Automaton total(const automaton::Automaton& automaton);

	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > total(const automaton::NFA < SymbolType, StateType > & automaton);

	template < class SymbolType, class StateType >
	static automaton::DFA < SymbolType, StateType > total(const automaton::DFA < SymbolType, StateType >& automaton);
};

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > Total::total(const automaton::NFA < SymbolType, StateType > & automaton) {
	if(! automaton.isDeterministic()) {
		throw exception::CommonException("Automaton must be deterministic to make its transition function total");
	}

	automaton::NFA < SymbolType, StateType > res(automaton);
	StateType nullState = common::createUnique(label::FailStateLabel::instance < StateType > ( ), automaton.getStates());
	res.addState(nullState);

	for(const auto& q : res.getStates()) {
		for(const auto& a : res.getInputAlphabet()) {
			if(res.getTransitions().find(ext::make_pair(q, a)) == res.getTransitions().end()) {
				res.addTransition(q, a, nullState);
			}
		}
	}

	return res;
}

template < class SymbolType, class StateType >
automaton::DFA < SymbolType, StateType > Total::total(const automaton::DFA < SymbolType, StateType > & automaton) {
	automaton::DFA < SymbolType, StateType > res(automaton);
	StateType nullState = common::createUnique(label::FailStateLabel::instance < StateType > ( ), automaton.getStates());
	res.addState(nullState);

	for(const auto& q : res.getStates()) {
		for(const auto& a : res.getInputAlphabet()) {
			if(res.getTransitions().find(ext::make_pair(q, a)) == res.getTransitions().end()) {
				res.addTransition(q, a, nullState);
			}
		}
	}

	return res;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* TOTAL_H_ */
