/*
 * AutomataUnionEpsilonTransition.h
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#ifndef AUTOMATA_UNION_EPSILON_H_
#define AUTOMATA_UNION_EPSILON_H_

#include <core/multipleDispatch.hpp>
#include <automaton/Automaton.h>
#include <automaton/FSM/EpsilonNFA.h>

namespace automaton {

namespace transform {

/**
 * Union two automata.
 *  - For finite automata A1, A2, we create automaton L accepting L(A1) \cup L(A2)
 */
class AutomataUnionEpsilonTransition : public alib::PromotingDoubleDispatch<AutomataUnionEpsilonTransition, automaton::Automaton, const automaton::AutomatonBase &> {
public:
	static automaton::Automaton unification(const automaton::Automaton& first, const automaton::Automaton& second);

	template < class SymbolType, class EpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > unification(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & first, const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & second);
	template < class SymbolType, class EpsilonType = DefaultEpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > unification(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second);
	template < class SymbolType, class EpsilonType = DefaultEpsilonType, class StateType >
	static automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > unification(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second);

};

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataUnionEpsilonTransition::unification(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & first, const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	ext::set<StateType> states;
	for(const auto& q : first.getStates())
		states.insert(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		states.insert(StateType(secondDefault, q));

	StateType q0 = common::createUnique(label::InitialStateLabel::instance < StateType > ( ), states);
	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(StateType(firstDefault, q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	res.addTransition(q0, StateType(firstDefault, first.getInitialState()));
	res.addTransition(q0, StateType(secondDefault, second.getInitialState()));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, q));

	return res;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataUnionEpsilonTransition::unification(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	ext::set<StateType> states;
	for(const auto& q : first.getStates())
		states.insert(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		states.insert(StateType(secondDefault, q));

	StateType q0 = common::createUnique(label::InitialStateLabel::instance < StateType > ( ), states);
	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(StateType(firstDefault, q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	res.addTransition(q0, StateType(firstDefault, first.getInitialState()));
	res.addTransition(q0, StateType(secondDefault, second.getInitialState()));

	for(const auto& t : first.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, q));

	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, q));

	return res;
}

template < class SymbolType, class EpsilonType, class StateType >
automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > AutomataUnionEpsilonTransition::unification(const automaton::DFA< SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	ext::set<StateType> states;
	for(const auto& q : first.getStates())
		states.insert(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		states.insert(StateType(secondDefault, q));

	StateType q0 = common::createUnique(label::InitialStateLabel::instance < StateType > ( ), states);
	automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& q : states)
		res.addState(q);

	for(const auto& q : first.getFinalStates())
		res.addFinalState(StateType(firstDefault, q));
	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));

	res.addTransition(q0, StateType(firstDefault, first.getInitialState()));
	res.addTransition(q0, StateType(secondDefault, second.getInitialState()));

	for(const auto& t : first.getTransitions())
		res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, t.second));

	for(const auto& t : second.getTransitions())
		res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, t.second));

	return res;
}

} /* namespace transform */

} /* namespace automaton */

#endif /* AUTOMATA_UNION_EPSILON_H_ */
