/*
 * ToAutomaton.cpp
 *
 * Created on: 1. 11. 2013
 * Author: Tomas Pecka
 */
#include "ToAutomaton.h"

#include <common/createUnique.hpp>

#include <label/InitialStateLabel.h>
#include <label/FinalStateLabel.h>
#include <registration/AlgoRegistration.hpp>

namespace grammar {

namespace convert {

automaton::Automaton ToAutomaton::convert(const grammar::Grammar& grammar) {
	 return dispatch(grammar.getData());
}

automaton::NFA < > ToAutomaton::convert(const grammar::LeftRG < > & grammar) {
	ext::map<DefaultSymbolType, DefaultStateType> stateMap;
	ext::set<DefaultStateType> states;

	// step 2
	for(const auto& symbol : grammar.getNonterminalAlphabet()) {
		DefaultStateType state ( symbol );
		states.insert(state);
		stateMap.insert(std::make_pair(symbol, state));
	}

	// step 1, 4
	DefaultStateType q0 = common::createUnique(label::InitialStateLabel::instance < DefaultStateType > ( ), states);
	states.insert(q0);
	automaton::NFA < > automaton(q0);
	automaton.setInputAlphabet(grammar.getTerminalAlphabet());
	automaton.setStates(states);

	// step 3
	for(const auto& rule : grammar.getRules()) {
		const DefaultSymbolType& lhs = rule.first;
		for(const auto& ruleRHS : rule.second) {
			if(ruleRHS.is<ext::pair<DefaultSymbolType, DefaultSymbolType>>()) { // if B->Ca => \delta(C,a)=B
				const ext::pair<DefaultSymbolType, DefaultSymbolType>& rhs = ruleRHS.get<ext::pair<DefaultSymbolType, DefaultSymbolType>>();
				automaton.addTransition(stateMap.find(rhs.first)->second, rhs.second, stateMap.find(lhs)->second);
			}
			else { // if B->a => \delta(StartState,a)=B
				const DefaultSymbolType& rhs = ruleRHS.get<DefaultSymbolType>();
				automaton.addTransition(q0, rhs, stateMap.find(lhs)->second);
			}
		}
	}

	// step 5
	automaton.addFinalState(stateMap.find(grammar.getInitialSymbol())->second);
	if(grammar.getGeneratesEpsilon())
		automaton.addFinalState(q0);

	return automaton;
}

auto ToAutomatonLeftRG = registration::OverloadRegister < ToAutomaton, automaton::NFA < >, grammar::LeftRG < > > ( ToAutomaton::convert );

automaton::NFA < > ToAutomaton::convert(const grammar::RightRG < > & grammar) {
	ext::map<DefaultSymbolType, DefaultStateType> stateMap;
	ext::set<DefaultStateType> states;

	// step2
	for(const auto& symbol : grammar.getNonterminalAlphabet()) {
		DefaultStateType state( symbol );
		states.insert(state);
		stateMap.insert(std::make_pair(symbol, state));
	}

	// step 1, 4
	DefaultStateType AState = common::createUnique(label::FinalStateLabel::instance < DefaultStateType > ( ), states);
	states.insert(AState);
	automaton::NFA < > automaton(stateMap.find(grammar.getInitialSymbol())->second);
	automaton.setStates(states);

	automaton.setInputAlphabet(grammar.getTerminalAlphabet());

	// step 3
	for(const auto& rule : grammar.getRules()) {
		const DefaultSymbolType& lhs = rule.first;
		for(const auto& ruleRHS : rule.second) {
			if(ruleRHS.is<ext::pair<DefaultSymbolType, DefaultSymbolType>>()) { // if B->aC => \delta(B,a)=C
				const ext::pair<DefaultSymbolType, DefaultSymbolType>& rhs = ruleRHS.get<ext::pair<DefaultSymbolType, DefaultSymbolType>>();
				automaton.addTransition(stateMap.find(lhs)->second, rhs.first, stateMap.find(rhs.second)->second);
			}
			else { // if B->a => \delta(B,a)=AState
				const DefaultSymbolType& rhs = ruleRHS.get<DefaultSymbolType>();
				automaton.addTransition(stateMap.find(lhs)->second, rhs, AState);
			}
		}
	}

	// step 5
	automaton.addFinalState(AState);
	if(grammar.getGeneratesEpsilon())
		automaton.addFinalState(automaton.getInitialState());

	return automaton;
}

auto ToAutomatonRightRG = registration::OverloadRegister < ToAutomaton, automaton::NFA < >, grammar::RightRG < > > ( ToAutomaton::convert );

template <class T>
automaton::NPDA < > ToAutomaton::convert(const T& grammar) {
	automaton::NPDA < > automaton(label::InitialStateLabel::instance < DefaultStateType > ( ), grammar.getInitialSymbol());

	automaton.setInputAlphabet(grammar.getTerminalAlphabet());

	automaton.setPushdownStoreAlphabet(grammar.getNonterminalAlphabet());
	for(const DefaultSymbolType& symbol : grammar.getTerminalAlphabet())
		automaton.addPushdownStoreSymbol(symbol);

	for(const std::pair<const DefaultSymbolType, ext::set<ext::vector<DefaultSymbolType>>>& kv : grammar.getRules())
		for(const ext::vector<DefaultSymbolType>& rhs : kv.second)
			automaton.addTransition(automaton.getInitialState(), ext::vector<DefaultSymbolType>{kv.first}, automaton.getInitialState(), rhs);

	for(const DefaultSymbolType& symbol : grammar.getTerminalAlphabet())
		automaton.addTransition(automaton.getInitialState(), symbol, ext::vector<DefaultSymbolType>{symbol}, automaton.getInitialState(), ext::vector<DefaultSymbolType>{});

	return automaton;
}

auto ToAutomatonCFG = registration::OverloadRegister < ToAutomaton, automaton::NPDA < >, grammar::CFG < > > ( ToAutomaton::convert );
auto ToAutomatonEpsilonFreeCFG = registration::OverloadRegister < ToAutomaton, automaton::NPDA < >, grammar::EpsilonFreeCFG < > > ( ToAutomaton::convert );

} /* namespace convert */

} /* namespace grammar */
