/*
 * ToRegExpAlgebraic.h
 *
 *  Created on: 4. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef GRAMMAR_TO_REG_EXP_ALGEBRAIC_H_
#define GRAMMAR_TO_REG_EXP_ALGEBRAIC_H_

#include <core/multipleDispatch.hpp>

#include <grammar/Grammar.h>
#include <grammar/Regular/RightRG.h>
#include <grammar/Regular/LeftRG.h>

#include <regexp/RegExp.h>

#include <equations/LeftRegularEquationSolver.h>
#include <equations/RightRegularEquationSolver.h>

#include <regexp/unbounded/UnboundedRegExp.h>
#include <regexp/unbounded/UnboundedRegExpElements.h>

namespace grammar {

namespace convert {

class ToRegExpAlgebraic : public alib::SingleDispatch<ToRegExpAlgebraic, regexp::RegExp, const grammar::GrammarBase &> {
public:
	/**
	 * @return regexp equivalent to source right regular grammar.
	 * @param grammar Grammar to convert
	 */
	static regexp::RegExp convert(const grammar::Grammar& grammar);

	template < class SymbolType >
	static regexp::UnboundedRegExp < SymbolType > convert(const grammar::RightRG < SymbolType > & grammar);
	template < class SymbolType >
	static regexp::UnboundedRegExp < SymbolType > convert(const grammar::LeftRG < SymbolType > & grammar);
};

template < class SymbolType >
regexp::UnboundedRegExp < SymbolType > ToRegExpAlgebraic::convert(const grammar::LeftRG < SymbolType > & grammar) {
	equations::LeftRegularEquationSolver < SymbolType > solver;

	for(const auto& symbol : grammar.getNonterminalAlphabet())
		solver.addSymbol(symbol);

	for(const auto & rule : grammar.getRules()) {
		const SymbolType& lhs = rule.first;

		for(const auto& ruleRHS : rule.second) {
			if(ruleRHS.template is<SymbolType>()) {
				const SymbolType& rhs = ruleRHS.template get<SymbolType>();
				solver.addEquation(lhs, regexp::UnboundedRegExpSymbol < SymbolType > (rhs));
			} else {
				const ext::pair<SymbolType, SymbolType>& rhs = ruleRHS.template get<ext::pair<SymbolType, SymbolType>>();
				solver.addEquation(lhs, rhs.first, regexp::UnboundedRegExpSymbol < SymbolType > (rhs.second));
			}
		}
	}
	if(grammar.getGeneratesEpsilon())
		solver.addEquation(grammar.getInitialSymbol(), regexp::UnboundedRegExpEpsilon < SymbolType > ());

	return solver.solve(grammar.getInitialSymbol());
}

template < class SymbolType >
regexp::UnboundedRegExp < SymbolType > ToRegExpAlgebraic::convert(const grammar::RightRG < SymbolType > & grammar) {
	equations::RightRegularEquationSolver < SymbolType > solver;

	for(const auto& symbol : grammar.getNonterminalAlphabet())
		solver.addSymbol(symbol);

	for(const auto & rule : grammar.getRules()) {
		const SymbolType& lhs = rule.first;

		for(const auto& ruleRHS : rule.second) {
			if(ruleRHS.template is<SymbolType>()) {
				const SymbolType& rhs = ruleRHS.template get<SymbolType>();
				solver.addEquation(lhs, regexp::UnboundedRegExpSymbol < SymbolType > (rhs));
			} else {
				const ext::pair<SymbolType, SymbolType>& rhs = ruleRHS.template get<ext::pair<SymbolType, SymbolType>>();
				solver.addEquation(lhs, rhs.second, regexp::UnboundedRegExpSymbol < SymbolType > (rhs.first));
			}
		}
	}
	if(grammar.getGeneratesEpsilon())
		solver.addEquation(grammar.getInitialSymbol(), regexp::UnboundedRegExpEpsilon < SymbolType > ());

	return solver.solve(grammar.getInitialSymbol());
}

} /* namespace covert */

} /* namespace grammar */

#endif /* GRAMMAR_TO_REG_EXP_ALGEBRAIC_H_ */
