/*
 * ToGNF.h
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#ifndef TO_GNF_H_
#define TO_GNF_H_

#include <core/multipleDispatch.hpp>
#include <map>
#include <algorithm>

#include <grammar/Grammar.h>

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include "EpsilonRemover.h"
#include "SimpleRulesRemover.h"
#include "LeftRecursionRemover.h"
#include <grammar/convert/ToGrammarRightRG.h>
#include <common/createUnique.hpp>

namespace grammar {

namespace simplify {

class ToGNF : public alib::SingleDispatch<ToGNF, grammar::Grammar, const grammar::GrammarBase &> {
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > assignNonterminals(const grammar::EpsilonFreeCFG < SymbolType > & origGrammar);
	template < class SymbolType >
	static grammar::GNF < SymbolType > convertInternal( const grammar::EpsilonFreeCFG < SymbolType > & origGrammar );

public:
	static grammar::Grammar convert( const grammar::Grammar & grammar );

	template < class SymbolType >
	static grammar::GNF < SymbolType > convert( const grammar::CFG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > convert( const grammar::EpsilonFreeCFG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > convert( const grammar::CNF < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > convert( const grammar::GNF < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > convert( const grammar::LG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > convert( const grammar::LeftLG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightRG < SymbolType > convert( const grammar::LeftRG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > convert( const grammar::RightLG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightRG < SymbolType > convert( const grammar::RightRG < SymbolType > & grammar );
};

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > ToGNF::assignNonterminals(const grammar::EpsilonFreeCFG < SymbolType > & origGrammar) {
	grammar::EpsilonFreeCFG < SymbolType > res(origGrammar.getInitialSymbol());
	res.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet());
	res.setTerminalAlphabet(origGrammar.getTerminalAlphabet());
	res.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	for(const std::pair<const SymbolType, ext::set<ext::vector<SymbolType>>>& rule : origGrammar.getRules()) {
		for(const ext::vector<SymbolType>& singleRHS : rule.second) {
			if(res.getTerminalAlphabet().count(singleRHS[0])) { //do not substitute terminals
				res.addRule(rule.first, singleRHS);
				continue;
			}
			const SymbolType& secondLHS = singleRHS[0];
			if(origGrammar.getRules().find(secondLHS) == origGrammar.getRules().end()) { //is there any right hand side to substitue with?
				//if not well this rule does not generate anything anyway
				continue;
			}

			for(const ext::vector<SymbolType>& secondSingleRHS : origGrammar.getRules().find(secondLHS)->second) { // do the substitution
				ext::vector<SymbolType> newRHS(secondSingleRHS);
				newRHS.insert(newRHS.end(), singleRHS.begin() + 1, singleRHS.end());
				res.addRule(rule.first, newRHS);
			}
		}
	}
	return res;
}

template < class SymbolType >
grammar::GNF < SymbolType > ToGNF::convertInternal( const grammar::EpsilonFreeCFG < SymbolType > & origGrammar ) {
	grammar::EpsilonFreeCFG < SymbolType > step(origGrammar);
	while(true) {
		grammar::EpsilonFreeCFG < SymbolType > nextStep = assignNonterminals(step);

		if(step == nextStep) break;
		step = std::move(nextStep);
	}

	grammar::GNF < SymbolType > res(step.getInitialSymbol());
	res.setTerminalAlphabet(step.getTerminalAlphabet());
	res.setNonterminalAlphabet(step.getNonterminalAlphabet());
	res.setGeneratesEpsilon(step.getGeneratesEpsilon());
	ext::map<SymbolType, SymbolType> terminalToPrimed;
	for(const SymbolType& terminal : step.getTerminalAlphabet()) {
		SymbolType primed = common::createUnique(terminal, res.getTerminalAlphabet(), res.getNonterminalAlphabet());
		terminalToPrimed.insert(std::make_pair(terminal, primed));
		res.addNonterminalSymbol(primed);
		res.addRule(primed, ext::make_pair(terminal, ext::vector<SymbolType> {}));
	}
	for(const auto& rule : step.getRules()) {
		for(const auto& rhs : rule.second) {
			ext::vector<SymbolType> convertedNonterminals;
			bool first = true;
			for(const SymbolType& rhsSymbol : rhs) {
				if(first) {
					first = false;
					continue;
				}

				if(res.getNonterminalAlphabet().count(rhsSymbol))
					convertedNonterminals.push_back(rhsSymbol);
				else
					convertedNonterminals.push_back(terminalToPrimed.find(rhsSymbol)->second);
			}
			res.addRule(rule.first, ext::make_pair(rhs[0], std::move(convertedNonterminals)));
		}
	}
	return res;
}

template < class SymbolType >
grammar::GNF < SymbolType > ToGNF::convert(const grammar::CFG < SymbolType > & origGrammar) {
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar))));
}

template < class SymbolType >
grammar::GNF < SymbolType > ToGNF::convert(const grammar::EpsilonFreeCFG < SymbolType > & origGrammar) {
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(origGrammar)));
}

template < class SymbolType >
grammar::GNF < SymbolType > ToGNF::convert(const grammar::CNF < SymbolType > & origGrammar) {
	return convertInternal(grammar::simplify::LeftRecursionRemover::remove(origGrammar));
}

template < class SymbolType >
grammar::GNF < SymbolType > ToGNF::convert(const grammar::GNF < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::GNF < SymbolType > ToGNF::convert(const grammar::LG < SymbolType > & origGrammar) {
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar))));
}

template < class SymbolType >
grammar::GNF < SymbolType > ToGNF::convert(const grammar::LeftLG < SymbolType > & origGrammar) {
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar))));
}

template < class SymbolType >
grammar::RightRG < SymbolType > ToGNF::convert(const grammar::LeftRG < SymbolType > & origGrammar) {
	return convert::ToGrammarRightRG::convert(origGrammar);
}

template < class SymbolType >
grammar::GNF < SymbolType > ToGNF::convert(const grammar::RightLG < SymbolType > & origGrammar) {
	return convertInternal(grammar::simplify::SimpleRulesRemover::remove(grammar::simplify::LeftRecursionRemover::remove(grammar::simplify::EpsilonRemover::remove(origGrammar))));
}

template < class SymbolType >
grammar::RightRG < SymbolType > ToGNF::convert(const grammar::RightRG < SymbolType > & origGrammar) {
	return origGrammar;
}

} /* namespace simplify */

} /* namespace grammar */

#endif /* TO_GNF_H_ */
