/*
 * Edmonds.cpp
 *
 *   Created on: Apr 7, 2016
 *       Author: Jan Broz
 */

#include "Edmonds.h"

#include <vector>
#include <list>
#include <queue>
#include <climits>
#include <stdexcept>

#include <graph/GraphClasses.h>
#include "../datastructs/BinomialHeap.h"
#include "../datastructs/FibonacciHeap.h"
#include "../datastructs/Components.h"
#include "../datastructs/Array.h"

#include <exception/CommonException.h>
#include <registration/AlgoRegistration.hpp>

namespace graph {

namespace spanningtree {

//====================================================================================================
//  data structures

/// wrapper for directed edge with additional info
struct Edge {
	DirectedEdge e;
	unsigned from;
	unsigned to;
	int weight;
	int origWeight;
	Edge * fParent;
	ext::vector<Edge*> fChildren;
	bool deleted;

	Edge( const DirectedEdge & edge, unsigned f, unsigned t, int w )
	 : e(edge), from(f), to(t), weight(w), origWeight(w), fParent(NULL), deleted(false) {}
};

static int compareEdges( Edge * const & e1, Edge * const & e2 )
{
	return e2->weight - e1->weight;
}

class EdgeQueue : public BinomialHeap<Edge*> {

 public:

	EdgeQueue() : BinomialHeap( compareEdges ), offset( 0 ) {}
	~EdgeQueue() {}

	void decreaseAll( int amount )  { offset -= amount; }

	Edge * extractMax()
	{
		if (size() == 0)
			return NULL;
		Edge * edge = BinomialHeap<Edge*>::extractMax();
		edge->weight += offset;
		return edge;
	}

 protected:

	int offset;

};

template< typename Type >
void vector_remove( ext::vector<Type> & vec, Type elem )
{
	for (auto it = vec.begin(); it != vec.end(); it++) {
		if (*it == elem) {
			vec.erase( it );
			break;
		}
	}
}

//====================================================================================================
//  algorithm

static AdjacencyListDirectedGraph edmonds_impl( const DirectedGraph & graph )
{
	AdjacencyListDirectedGraph spanningTree;    // spanning tree

	unsigned nodeCnt = graph.nodeCount();
	unsigned edgeCnt = graph.edgeCount();

	ext::list<Edge> edges;       // just to ease deallocation
	std::queue<unsigned> roots;                  ///< silne komponenty, ktere jsou koreny v aktualnim grafu (nevede do nich zadny uzel)
	Array<EdgeQueue> queues( nodeCnt );      ///< vstupni hrany do danych uzlu (razene dle priority)
	Array<Edge*> enter( nodeCnt );           ///< vstupni hrany do silnych komponent
	Array<ext::vector<Edge*>> h( nodeCnt );  ///< hrany, ktere mohou byt v kostre
	ext::vector<unsigned> rset;                       ///< rooti vyslednych stromu Tarjanovy implementace
	Components2 strong( nodeCnt );           ///< silne komponenty (reprezentanti)
	Components2 weak( nodeCnt );             ///< slabe komponenty (reprezentanti)
	Array<unsigned> min( nodeCnt );              ///< urcuje finalni korenove uzly orientovanych koster
	ext::vector<Edge*> fRoots;               ///< koreny lesa F
	Array<ext::vector<Edge*>> cycles( nodeCnt );  ///< nalezene cykly
	Array<Edge*> lambda( nodeCnt );          ///< listy lesa F

	// this algorithm works on number identification of nodes and edges
	ext::vector<Node> id2node( nodeCnt );
	ext::vector<DirectedEdge> id2edge( edgeCnt );
	ext::unordered_map<Node,unsigned> node2id;
	ext::unordered_map<DirectedEdge,unsigned> edge2id;
	unsigned id = 0;
	for (const Node & node : graph.getNodes()) {
		id2node[ id ] = node;
		node2id[ node ] = id;
		id++;
	}
/*	id = 0;
	for (const DirectedEdge & edge : graph.getEdges()) {
		id2edge[ id ] = edge;
		edge2id[ edge ] = id;
		id++;
	}
*/
	// fill up data structures
	for (unsigned nodeID = 0; nodeID < nodeCnt; nodeID++) {
		roots.push( nodeID );
		enter[ nodeID ] = NULL;
		min[ nodeID ] = nodeID;
		strong.MakeSet( nodeID );
		weak.MakeSet( nodeID );
		lambda[ nodeID ] = NULL;
	}
	for (const DirectedEdge & edge : graph.getEdges()) {
		int weight;
		try {
			weight = graph.getEdgeValue( edge );
		} catch (const std::out_of_range & ex) {
			throw exception::CommonException("Edmonds: an edge is missing a capacity");
		}
		edges.emplace_back( edge, node2id[ edge.getFromNode() ], node2id[ edge.getToNode() ], weight );
		queues[ node2id[edge.getToNode()] ].insert( &edges.back() );
	}

	while (roots.size() != 0) {
		unsigned root = roots.front(); roots.pop();
		Edge * edge = queues[ root ].extractMax();
		if (edge != NULL) {
			if (cycles[ strong.Find( root ) ].size() <= 1) {
				lambda[ root ] = edge;
			} else if (strong.Find( edge->from ) != strong.Find( edge->to )) {
				for (Edge * e : cycles[ strong.Find( root ) ]) {
					vector_remove( fRoots, e );   //je v cyklu, nemuze byt rootem
					edge->fChildren.push_back( e );
					e->fParent = edge;
				}
			}
		}
		if (edge == NULL) {
			rset.push_back( root ); //pak uzel nema vstupni hranu (je root)
		} else if (strong.Find( edge->from ) == root) { //nalezli jsme hranu v ramci silne komponenty
			roots.push( root ); //uzel vratimem do fronty
		} else {
			h[ edge->from ].push_back( edge ); //tato hrana je adeptem na hranu MST
			fRoots.push_back( edge ); //tato hrana je korenem v lese F
			if (weak.Find( edge->from ) != weak.Find( edge->to )) { //pocatecni a koncovy uzel jsou v ruznych slabych komponentach
				weak.Union( weak.Find( edge->from ), weak.Find( edge->to ));
				enter[ root ] = edge; //edge je vstupni hranou do root
			} else { //pocatecni a koncovy uzel jsou ve stejne slabe komponente
				Edge * cycleEdge = edge;
				int maxVal = INT_MIN;
				int vertex = INT_MIN;

				ext::vector<Edge*> & cycle = cycles[ strong.Find( root ) ];
				cycle.clear();
				while (cycleEdge != NULL) { //hledame hranu v cyklu s maximalni vahou (ktera nejvice poskodi cyklus)
					cycle.push_back( cycleEdge );
					if (cycleEdge->weight > maxVal) {
						maxVal = cycleEdge->weight;
						vertex = strong.Find( cycleEdge->to );
					}
					cycleEdge = enter[ strong.Find( cycleEdge->from ) ];
				}
				queues[ root ].decreaseAll( edge->weight - maxVal );
				min[ root ] = min[ vertex ];

				//collapse cycle
				cycleEdge = enter[ strong.Find( edge->from ) ];
				while (cycleEdge != NULL) {
					queues[ strong.Find( cycleEdge->to ) ].decreaseAll( cycleEdge->weight - maxVal ); //pridame prichozim hranam hodnotu, ktera znaci, o kolik se vypustenim dane cyklove hrany a pridanim dane necyklove zhorsi reseni
					queues[ root ].mergeWith( std::move( queues[ strong.Find( cycleEdge->to ) ] ) );
					strong.Union( root, strong.Find( cycleEdge->to ) );
					cycleEdge = enter[ strong.Find( cycleEdge->from ) ];
				}
				roots.push( root );
			}
		}
	}


	ext::vector<Edge*> b;
	ext::vector<unsigned> r;
	unsigned fRootIndex = 0;

	while (!rset.empty() || fRootIndex < fRoots.size()) {
		unsigned v = 0;
		if (!rset.empty()) {
			v = min[ rset[0] ]; rset.erase( rset.begin() );
			r.push_back(v);
		} else {
			Edge * e = fRoots[ fRootIndex ];
			if (e->deleted) {
				fRootIndex++;
				continue;
			}
			v = e->to;
			b.push_back(e);
			fRootIndex++;
		}
		Edge * curr = lambda[ v ];
		while (curr != NULL && !curr->deleted) {
			curr->deleted = true;
			fRoots.insert( fRoots.begin(), curr->fChildren.begin(), curr->fChildren.end() );
			curr = curr->fParent;
		}
	}

	for (const Node & node : graph.getNodes())
		spanningTree.addNode( node );
	for (Edge * edge : b)
		spanningTree.addEdge( edge->e, edge->origWeight );

	return spanningTree;
}

Graph Edmonds::edmonds( const Graph & graph )
{
	return dispatch( graph.getData() );
}

AdjacencyListDirectedGraph Edmonds::edmonds( const DirectedGraph & graph )
{
	return edmonds_impl( graph );
}

auto EdmondsDirectedGraph = registration::OverloadRegister < Edmonds,  graph::AdjacencyListDirectedGraph, graph::DirectedGraph > ( Edmonds::edmonds );

} // spanningtree

} // graph
