/*
 * List.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_LIST_H_
#define OBJECTS_LIST_H_

#include <list>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"
#include <object/AnyObject.h>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class ElementType >
class ObjectsList final : public ext::list < ElementType >, public ContainerBase {
public:
	explicit ObjectsList ( ext::list < ElementType > );
	explicit ObjectsList ( );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsList & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "List";

		return xmlTagName;
	}

	static ObjectsList parse ( ext::deque < sax::Token >::iterator & input );
	static ext::list < ElementType > parseRaw ( ext::deque < sax::Token >::iterator & input );

	void compose ( ext::deque < sax::Token > & out ) const;
	static void compose ( ext::deque < sax::Token > & out, const ext::list < ElementType > & input );

	virtual alib::ObjectBase * inc ( ) &&;

	static ext::list < alib::Object > normalizeRaw ( ext::list < ElementType > && source ) {
		ext::list < alib::Object > res;
		for ( ElementType & element : source ) {
			res.push_back ( alib::Object ( alib::AnyObject < ElementType > ( std::move ( element ) ) ) );
		}
		return res;
	}

	typedef ObjectsList < > normalized_type;

	virtual ContainerBase * normalize ( ) && {
		return new ObjectsList < alib::Object > ( normalizeRaw ( std::move ( * this ) ) );
	}
};

template < class ElementType >
ObjectsList < ElementType >::ObjectsList ( ext::list < ElementType > raw ) : ext::list < ElementType > ( std::move ( raw ) ) {
}

template < class ElementType >
ObjectsList < ElementType >::ObjectsList ( ) : ext::list < ElementType > ( ) {
}

template < class ElementType >
ContainerBase * ObjectsList < ElementType >::clone ( ) const {
	return new ObjectsList ( * this );
}

template < class ElementType >
ContainerBase * ObjectsList < ElementType >::plunder ( ) && {
	return new ObjectsList ( std::move ( * this ) );
}

template < class ElementType >
int ObjectsList < ElementType >::compare ( const ObjectsList & other ) const {
	static ext::compare < ext::list < ElementType > > comp;

	return comp ( static_cast < const ext::list < ElementType > & > ( * this ), static_cast < const ext::list < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsList < ElementType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsList " << static_cast < const ext::list < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsList < ElementType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class ElementType >
ObjectsList < ElementType > ObjectsList < ElementType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return ObjectsList < ElementType > ( parseRaw ( input ) );
}

template < class ElementType >
void ObjectsList < ElementType >::compose ( ext::deque < sax::Token > & out ) const {
	compose ( out, * this );
}

template < class ElementType >
ext::list < ElementType > ObjectsList < ElementType >::parseRaw ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsList < >::getXmlTagName() );

	ext::list < ElementType > list;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		list.push_back ( alib::xmlApi < ElementType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsList < >::getXmlTagName() );
	return list;
}

template < class ElementType >
void ObjectsList < ElementType >::compose ( ext::deque < sax::Token > & out, const ext::list < ElementType > & input ) {
	out.emplace_back ( ObjectsList < >::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	for ( const ElementType & item : input )
		alib::xmlApi < ElementType >::compose ( out, item );

	out.emplace_back ( ObjectsList < >::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class ElementType >
alib::ObjectBase* ObjectsList < ElementType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < ext::list < T > > {
	static ext::list < T > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static std::string xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const ext::list < T > & data );
};

template < typename T >
ext::list < T > xmlApi < ext::list < T > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return container::ObjectsList < T >::parseRaw ( input );
}

template < typename T >
bool xmlApi < ext::list < T > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName() );
}

template < typename T >
std::string xmlApi < ext::list < T > >::xmlTagName ( ) {
	return container::ObjectsList < >::getXmlTagName();
}

template < typename T >
void xmlApi < ext::list < T > >::compose ( ext::deque < sax::Token > & output, const ext::list < T > & input ) {
	return container::ObjectsList < T >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_LIST_H_ */
