/*
 * Set.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_SET_H_
#define OBJECTS_SET_H_

#include <set>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"
#include <object/AnyObject.h>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class ElementType >
class ObjectsSet final : public ext::set < ElementType >, public ContainerBase {
public:
	explicit ObjectsSet ( ext::set < ElementType > );
	explicit ObjectsSet ( );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsSet & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Set";

		return xmlTagName;
	}

	static ObjectsSet parse ( ext::deque < sax::Token >::iterator & input );
	static ext::set < ElementType > parseRaw ( ext::deque < sax::Token >::iterator & input );

	void compose ( ext::deque < sax::Token > & out ) const;
	static void compose ( ext::deque < sax::Token > & out, const ext::set < ElementType > & input );

	virtual alib::ObjectBase * inc ( ) &&;

	static ext::set < alib::Object > normalizeRaw ( ext::set < ElementType > && source ) {
		ext::set < alib::Object > res;
		for ( ElementType && element : ext::make_moveable_set ( source ) ) {
			res.insert ( alib::Object ( alib::AnyObject < ElementType > ( std::move ( element ) ) ) );
		}
		return res;
	}

	typedef ObjectsSet < > normalized_type;

	virtual ContainerBase * normalize ( ) && {
		return new ObjectsSet < alib::Object > ( normalizeRaw ( std::move ( * this ) ) );
	}
};

template < class ElementType >
ObjectsSet < ElementType >::ObjectsSet(ext::set<ElementType> raw) : ext::set<ElementType>(std::move(raw)) {

}

template < class ElementType >
ObjectsSet < ElementType >::ObjectsSet() : ext::set<ElementType>() {

}

template < class ElementType >
ContainerBase* ObjectsSet < ElementType >::clone() const {
	return new ObjectsSet(*this);
}

template < class ElementType >
ContainerBase* ObjectsSet < ElementType >::plunder() && {
	return new ObjectsSet(std::move(*this));
}

template < class ElementType >
int ObjectsSet < ElementType >::compare(const ObjectsSet& other) const {
	static ext::compare<ext::set<ElementType>> comp;
	return comp ( static_cast < const ext::set < ElementType > & > ( * this ), static_cast < const ext::set < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsSet < ElementType >::operator>>(std::ostream& os) const {
	os << "(ObjectsSet " << static_cast < const ext::set < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsSet < ElementType >::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class ElementType >
ObjectsSet < ElementType > ObjectsSet < ElementType >::parse(ext::deque<sax::Token>::iterator& input) {
	return ObjectsSet < ElementType >(parseRaw(input));
}

template < class ElementType >
void ObjectsSet < ElementType >::compose(ext::deque<sax::Token>& out) const {
	compose(out, *this);
}

template < class ElementType >
ext::set < ElementType > ObjectsSet < ElementType >::parseRaw ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsSet < >::getXmlTagName() );

	ext::set < ElementType > set;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		set.insert ( alib::xmlApi < ElementType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsSet < >::getXmlTagName() );
	return set;
}

template < class ElementType >
void ObjectsSet < ElementType >::compose ( ext::deque < sax::Token > & out, const ext::set < ElementType > & input ) {
	out.emplace_back ( ObjectsSet < >::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	for ( const ElementType & item : input )
		alib::xmlApi < ElementType >::compose ( out, item );

	out.emplace_back ( ObjectsSet < >::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class ElementType >
alib::ObjectBase* ObjectsSet < ElementType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < ext::set < T > > {
	static ext::set < T > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static std::string xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const ext::set < T > & data );
};

template < typename T >
ext::set < T > xmlApi < ext::set < T > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return container::ObjectsSet < T >::parseRaw ( input );
}

template < typename T >
bool xmlApi < ext::set < T > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName() );
}

template < typename T >
std::string xmlApi < ext::set < T > >::xmlTagName ( ) {
	return container::ObjectsSet < >::getXmlTagName();
}

template < typename T >
void xmlApi < ext::set < T > >::compose ( ext::deque < sax::Token > & output, const ext::set < T > & input ) {
	return container::ObjectsSet < T >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_SET_H_ */
