/*
 * ObjectsTree.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_TREE_H_
#define OBJECTS_TREE_H_

#include <tree>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"
#include <object/AnyObject.h>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class ElementType >
class ObjectsTree final : public ext::tree < ElementType >, public ContainerBase {
public:
	explicit ObjectsTree ( ext::tree < ElementType > );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsTree & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "Tree";

		return xmlTagName;
	}

	static ObjectsTree parse ( ext::deque < sax::Token >::iterator & input );
	static ext::tree < ElementType > parseRaw ( ext::deque < sax::Token >::iterator & input );

	void compose ( ext::deque < sax::Token > & out ) const;
	static void compose ( ext::deque < sax::Token > & out, const ext::tree < ElementType > & input );

	virtual alib::ObjectBase * inc ( ) &&;

	ext::tree < alib::Object > normalizeRaw ( ext::tree < ElementType > && node ) {
		ext::vector < ext::tree < alib::Object > > children;

		for ( ext::tree < ElementType > & child : node.getChildren ( ) ) {
			children.push_back ( normalizeRaw ( std::move ( child ) ) );
		}

		return ext::tree < alib::Object > ( alib::Object ( alib::AnyObject < ElementType > ( std::move ( node.getData ( ) ) ) ), std::move ( children ) );
	}

	typedef ObjectsTree < > normalized_type;

	virtual ContainerBase * normalize ( ) && {
		return new ObjectsTree < alib::Object > ( normalizeRaw ( std::move ( * this ) ) );
	}
};

template < class ElementType >
ObjectsTree < ElementType >::ObjectsTree ( ext::tree < ElementType > raw ) : ext::tree < ElementType > ( std::move ( raw ) ) {
}

template < class ElementType >
ContainerBase * ObjectsTree < ElementType >::clone ( ) const {
	return new ObjectsTree ( * this );
}

template < class ElementType >
ContainerBase * ObjectsTree < ElementType >::plunder ( ) && {
	return new ObjectsTree ( std::move ( * this ) );
}

template < class ElementType >
int ObjectsTree < ElementType >::compare ( const ObjectsTree & other ) const {
	static ext::compare < ext::tree < ElementType > > comp;

	return comp ( static_cast < const ext::tree < ElementType > & > ( * this ), static_cast < const ext::tree < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsTree < ElementType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsTree " << static_cast < const ext::tree < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsTree < ElementType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class ElementType >
ObjectsTree <ElementType > ObjectsTree < ElementType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return ObjectsTree < ElementType > ( parseRaw ( input ) );
}

template < class ElementType >
void ObjectsTree < ElementType >::compose ( ext::deque < sax::Token > & out ) const {
	compose ( out, * this );
}

template < class ElementType >
ext::tree < ElementType > ObjectsTree < ElementType >::parseRaw ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsTree < >::getXmlTagName ( ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Root" );

	ext::tree < ElementType > tree ( alib::xmlApi < ElementType >::parse ( input ) );

	typename ext::tree < ElementType >::const_children_iterator under = tree.root ( );

	unsigned level = 0;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) ) {
		if ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, "Children" ) ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Children" );

			if ( level > 0 )
				under = std::prev ( under->end ( ) );

			++level;
		}

		tree.push_back ( under, alib::xmlApi < ElementType >::parse ( input ) );

		while ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::END_ELEMENT, "Children" ) ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Children" );

			under = under->parent ( );
			--level;
		}
	}

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Root" );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsTree < >::getXmlTagName ( ) );
	return tree;
}

template < class ElementType >
void ObjectsTree < ElementType >::compose ( ext::deque < sax::Token > & out, const ext::tree < ElementType > & input ) {
	out.emplace_back ( ObjectsTree < >::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	out.emplace_back ( "Root", sax::Token::TokenType::START_ELEMENT );

	unsigned level = 0;

	for ( typename ext::tree < ElementType >::const_prefix_iterator iter = input.prefix_begin ( ); iter != input.prefix_end ( ); ) {
		while ( iter.getLevel ( ) > level ) {
			out.emplace_back ( "Children", sax::Token::TokenType::START_ELEMENT );
			++level;
		}

		alib::xmlApi < ElementType >::compose ( out, * iter );
		++iter;

		while ( iter.getLevel ( ) < level ) {
			out.emplace_back ( "Children", sax::Token::TokenType::END_ELEMENT );
			--level;
		}
	}

	out.emplace_back ( "Root", sax::Token::TokenType::END_ELEMENT );
	out.emplace_back ( ObjectsTree < >::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < class ElementType >
alib::ObjectBase* ObjectsTree < ElementType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < ext::tree < T > > {
	static ext::tree < T > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static std::string xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const ext::tree < T > & data );
};

template < typename T >
ext::tree < T > xmlApi < ext::tree < T > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return container::ObjectsTree < T >::parseRaw ( input );
}

template < typename T >
bool xmlApi < ext::tree < T > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T >
std::string xmlApi < ext::tree < T > >::xmlTagName ( ) {
	return container::ObjectsTree < >::getXmlTagName ( );
}

template < typename T >
void xmlApi < ext::tree < T > >::compose ( ext::deque < sax::Token > & output, const ext::tree < T > & input ) {
	return container::ObjectsTree < T >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_TREE_H_ */
