/*
 * Author: Radovan Cerveny
 */

#include "../sax/FromXMLParserHelper.h"
#include "MeasurementResultsXml.hpp"
#include "../exception/CommonException.h"

namespace measurements {

ext::deque < sax::Token > MeasurementResultsXml::compose ( const MeasurementResults & mr ) {
	ext::deque < sax::Token > tokens;
	compose ( mr, tokens );
	return tokens;
}

void MeasurementResultsXml::compose ( const MeasurementResults & mr, ext::deque < sax::Token > & tokens ) {
	tokens.emplace_back ( MEASUREMENT_RESULTS_TAG, sax::Token::TokenType::START_ELEMENT );

	MeasurementResultsXml::composeMeasurementFrames ( tokens, 0, mr.frames );

	tokens.emplace_back ( MEASUREMENT_RESULTS_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementResultsXml::composeMeasurementFrames ( ext::deque < sax::Token > & tokens, unsigned idx, const stealth_vector < MeasurementFrame > & frames ) {

	const MeasurementFrame & frame = frames[idx];

	tokens.emplace_back ( MEASUREMENT_FRAME_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( MEASUREMENT_FRAME_NAME_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( measurements::to_string ( frame.name ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( MEASUREMENT_FRAME_NAME_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( MEASUREMENT_FRAME_TYPE_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( measurements::to_string ( frame.type ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( MEASUREMENT_FRAME_TYPE_TAG, sax::Token::TokenType::END_ELEMENT );

	if ( frame.type != measurements::Type::ROOT ) {
		composeTimeDataFrame ( tokens, idx, frames );
		composeMemoryDataFrame ( tokens, idx, frames );
		composeCounterDataFrame ( tokens, idx, frames );
	}

	tokens.emplace_back ( MEASUREMENT_SUBFRAMES_TAG, sax::Token::TokenType::START_ELEMENT );

	for ( unsigned subIdx : frame.subIdxs )
		composeMeasurementFrames ( tokens, subIdx, frames );

	tokens.emplace_back ( MEASUREMENT_SUBFRAMES_TAG, sax::Token::TokenType::END_ELEMENT );

	tokens.emplace_back ( MEASUREMENT_FRAME_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementResultsXml::composeTimeDataFrame ( ext::deque < sax::Token > & tokens, unsigned idx, const stealth_vector < MeasurementFrame > & frames ) {
	const TimeDataFrame & frame = frames[idx].time;

	tokens.emplace_back ( TIME_DATA_FRAME_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( TIME_DATA_FRAME_DURATION_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( ext::to_string ( frame.duration.count ( ) ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( TIME_DATA_FRAME_DURATION_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( TIME_DATA_FRAME_IN_FRAME_DURATION_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( ext::to_string ( frame.inFrameDuration.count ( ) ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( TIME_DATA_FRAME_IN_FRAME_DURATION_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( TIME_DATA_FRAME_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementResultsXml::composeMemoryDataFrame ( ext::deque < sax::Token > & tokens, unsigned idx, const stealth_vector < MeasurementFrame > & frames ) {
	const MemoryDataFrame & frame = frames[idx].memory;

	tokens.emplace_back ( MEMORY_DATA_FRAME_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( MEMORY_DATA_FRAME_START_HEAP_USAGE_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( ext::to_string ( frame.startHeapUsage ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( MEMORY_DATA_FRAME_START_HEAP_USAGE_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( MEMORY_DATA_FRAME_END_HEAP_USAGE_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( ext::to_string ( frame.endHeapUsage ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( MEMORY_DATA_FRAME_END_HEAP_USAGE_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( MEMORY_DATA_FRAME_HIGH_WATERMARK_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( ext::to_string ( frame.highWatermark ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( MEMORY_DATA_FRAME_HIGH_WATERMARK_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( MEMORY_DATA_FRAME_IN_FRAME_HIGH_WATERMARK_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( ext::to_string ( frame.inFrameHighWatermark ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( MEMORY_DATA_FRAME_IN_FRAME_HIGH_WATERMARK_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( MEMORY_DATA_FRAME_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementResultsXml::composeCounterDataFrame ( ext::deque < sax::Token > & tokens, unsigned idx, const stealth_vector < MeasurementFrame > & frames ) {
	const CounterDataFrame & cdf = frames[idx].counter;

	tokens.emplace_back ( COUNTER_DATA_FRAME_TAG, sax::Token::TokenType::START_ELEMENT );

	tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTERS_TAG, sax::Token::TokenType::START_ELEMENT );

	for ( const auto & elem : cdf.counters ) {
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_TAG, sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_NAME_TAG, sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( measurements::to_string ( elem.first ), sax::Token::TokenType::CHARACTER );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_NAME_TAG, sax::Token::TokenType::END_ELEMENT );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_VALUE_TAG, sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( ext::to_string ( elem.second ), sax::Token::TokenType::CHARACTER );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_VALUE_TAG, sax::Token::TokenType::END_ELEMENT );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_TAG, sax::Token::TokenType::END_ELEMENT );
	}

	tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTERS_TAG, sax::Token::TokenType::END_ELEMENT );

	tokens.emplace_back ( COUNTER_DATA_FRAME_IN_FRAME_COUNTERS_TAG, sax::Token::TokenType::START_ELEMENT );

	for ( const auto & elem : cdf.inFrameCounters ) {
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_TAG, sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_NAME_TAG, sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( measurements::to_string ( elem.first ), sax::Token::TokenType::CHARACTER );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_NAME_TAG, sax::Token::TokenType::END_ELEMENT );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_VALUE_TAG, sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( ext::to_string ( elem.second ), sax::Token::TokenType::CHARACTER );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_VALUE_TAG, sax::Token::TokenType::END_ELEMENT );
		tokens.emplace_back ( COUNTER_DATA_FRAME_COUNTER_TAG, sax::Token::TokenType::END_ELEMENT );
	}

	tokens.emplace_back ( COUNTER_DATA_FRAME_IN_FRAME_COUNTERS_TAG, sax::Token::TokenType::END_ELEMENT );

	tokens.emplace_back ( COUNTER_DATA_FRAME_TAG, sax::Token::TokenType::END_ELEMENT );
}

MeasurementResults MeasurementResultsXml::parse ( ext::deque < sax::Token > & tokens ) {
	MeasurementResults mr;

	ext::deque < sax::Token >::iterator iter = tokens.begin ( );

	if ( iter == tokens.end ( ) ) throw::exception::CommonException ( "Empty tokens list" );

	mr = parse ( iter );

	if ( iter != tokens.end ( ) ) throw::exception::CommonException ( "Unexpected tokens at the end of the xml" );

	return mr;
}

MeasurementResults MeasurementResultsXml::parse ( ext::deque < sax::Token >::iterator & iter ) {
	MeasurementResults mr;

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_RESULTS_TAG );

	parseRootMeasurementFrame ( mr, iter );

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_RESULTS_TAG );

	return mr;
}

void MeasurementResultsXml::parseRootMeasurementFrame ( MeasurementResults & mr, ext::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_FRAME_TAG );

	mr.frames.emplace_back ( "Root", measurements::Type::ROOT, 0 );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == MEASUREMENT_FRAME_NAME_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_FRAME_NAME_TAG );
			sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_FRAME_NAME_TAG );
		} else if ( data == MEASUREMENT_FRAME_TYPE_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_FRAME_TYPE_TAG );
			sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_FRAME_TYPE_TAG );
		} else if ( data == MEASUREMENT_SUBFRAMES_TAG ) {

			parseSubframes ( 0, mr, iter );
		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_FRAME_TAG );
}

void MeasurementResultsXml::parseSubframes ( unsigned parentIdx, MeasurementResults & mr, ext::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_SUBFRAMES_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == MEASUREMENT_FRAME_TAG )
			parseMeasurementFrame ( parentIdx, mr, iter );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_SUBFRAMES_TAG );
}

void MeasurementResultsXml::parseMeasurementFrame ( unsigned parentIdx, MeasurementResults & mr, ext::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_FRAME_TAG );

	unsigned mfIdx = mr.frames.size ( );

	mr.frames.emplace_back ( "Root", measurements::Type::ROOT, parentIdx );
	MeasurementFrame & mf = mr.frames.back ( );

	mr.frames[parentIdx].subIdxs.push_back ( mfIdx );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == MEASUREMENT_FRAME_NAME_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_FRAME_NAME_TAG );
			mf.name = measurements::stealthStringFromString ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_FRAME_NAME_TAG );
		} else if ( data == MEASUREMENT_FRAME_TYPE_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_FRAME_TYPE_TAG );
			mf.type = measurements::measurementTypeFromString ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_FRAME_TYPE_TAG );
		} else if ( data == TIME_DATA_FRAME_TAG ) {
			parseTimeDataFrame ( mf, iter );
		} else if ( data == MEMORY_DATA_FRAME_TAG ) {
			parseMemoryDataFrame ( mf, iter );
		} else if ( data == COUNTER_DATA_FRAME_TAG ) {
			parseCounterDataFrame ( mf, iter );
		} else if ( data == MEASUREMENT_SUBFRAMES_TAG ) {
			parseSubframes ( mfIdx, mr, iter );
		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_FRAME_TAG );
}

void MeasurementResultsXml::parseTimeDataFrame ( MeasurementFrame & mf, ext::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, TIME_DATA_FRAME_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == TIME_DATA_FRAME_DURATION_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, TIME_DATA_FRAME_DURATION_TAG );
			mf.time.duration = TimeDataFrame::value_type ( valueTypeFromString < TimeDataFrame::value_type::rep > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, TIME_DATA_FRAME_DURATION_TAG );
		} else if ( data == TIME_DATA_FRAME_IN_FRAME_DURATION_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, TIME_DATA_FRAME_IN_FRAME_DURATION_TAG );
			mf.time.inFrameDuration = TimeDataFrame::value_type ( valueTypeFromString < TimeDataFrame::value_type::rep > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, TIME_DATA_FRAME_IN_FRAME_DURATION_TAG );
		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, TIME_DATA_FRAME_TAG );
}

void MeasurementResultsXml::parseMemoryDataFrame ( MeasurementFrame & mf, ext::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEMORY_DATA_FRAME_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == MEMORY_DATA_FRAME_START_HEAP_USAGE_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEMORY_DATA_FRAME_START_HEAP_USAGE_TAG );
			mf.memory.startHeapUsage = valueTypeFromString < MemoryDataFrame::value_type > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEMORY_DATA_FRAME_START_HEAP_USAGE_TAG );
		} else if ( data == MEMORY_DATA_FRAME_END_HEAP_USAGE_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEMORY_DATA_FRAME_END_HEAP_USAGE_TAG );
			mf.memory.endHeapUsage = valueTypeFromString < MemoryDataFrame::value_type > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEMORY_DATA_FRAME_END_HEAP_USAGE_TAG );
		} else if ( data == MEMORY_DATA_FRAME_HIGH_WATERMARK_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEMORY_DATA_FRAME_HIGH_WATERMARK_TAG );
			mf.memory.highWatermark = valueTypeFromString < MemoryDataFrame::value_type > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEMORY_DATA_FRAME_HIGH_WATERMARK_TAG );
		} else if ( data == MEMORY_DATA_FRAME_IN_FRAME_HIGH_WATERMARK_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEMORY_DATA_FRAME_IN_FRAME_HIGH_WATERMARK_TAG );
			mf.memory.inFrameHighWatermark = valueTypeFromString < MemoryDataFrame::value_type > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEMORY_DATA_FRAME_IN_FRAME_HIGH_WATERMARK_TAG );
		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEMORY_DATA_FRAME_TAG );
}

void MeasurementResultsXml::parseCounterDataFrame ( MeasurementFrame & mf, ext::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, COUNTER_DATA_FRAME_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == COUNTER_DATA_FRAME_COUNTERS_TAG )
			parseCounterDataFrameCounters ( COUNTER_DATA_FRAME_COUNTERS_TAG, mf.counter.counters, iter );
		else if ( data == COUNTER_DATA_FRAME_IN_FRAME_COUNTERS_TAG )
			parseCounterDataFrameCounters ( COUNTER_DATA_FRAME_IN_FRAME_COUNTERS_TAG, mf.counter.inFrameCounters, iter );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, COUNTER_DATA_FRAME_TAG );
}

void MeasurementResultsXml::parseCounterDataFrameCounters ( const std::string & tag, stealth_map < stealth_string, CounterHint::value_type > & counters, ext::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, tag );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == COUNTER_DATA_FRAME_COUNTER_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, COUNTER_DATA_FRAME_COUNTER_TAG );

			stealth_string counterName;
			CounterHint::value_type counterValue = { };

			for ( ; ; ) {
				if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

				std::string counterData = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

				if ( counterData == COUNTER_DATA_FRAME_COUNTER_NAME_TAG ) {
					sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, COUNTER_DATA_FRAME_COUNTER_NAME_TAG );
					counterName = stealthStringFromString ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
					sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, COUNTER_DATA_FRAME_COUNTER_NAME_TAG );
				} else if ( counterData == COUNTER_DATA_FRAME_COUNTER_VALUE_TAG ) {
					sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, COUNTER_DATA_FRAME_COUNTER_VALUE_TAG );
					counterValue = valueTypeFromString < CounterHint::value_type > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
					sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, COUNTER_DATA_FRAME_COUNTER_VALUE_TAG );
				} else {
					break;
				}
			}

			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, COUNTER_DATA_FRAME_COUNTER_TAG );

			counters[counterName] = counterValue;

		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, tag );
}

template < typename T >
T MeasurementResultsXml::valueTypeFromString ( const std::string & str ) {
	std::istringstream iss ( str );
	T val;

	iss >> val;
	return val;
}

std::string MeasurementResultsXml::MEASUREMENT_RESULTS_TAG = "MeasurementResults";

std::string MeasurementResultsXml::MEASUREMENT_FRAME_TAG = "MeasurementFrame";
std::string MeasurementResultsXml::MEASUREMENT_FRAME_NAME_TAG = "Name";
std::string MeasurementResultsXml::MEASUREMENT_FRAME_TYPE_TAG = "Type";

std::string MeasurementResultsXml::MEASUREMENT_SUBFRAMES_TAG = "SubFrames";

std::string MeasurementResultsXml::TIME_DATA_FRAME_TAG = "TimeData";
std::string MeasurementResultsXml::TIME_DATA_FRAME_DURATION_TAG = "Duration";
std::string MeasurementResultsXml::TIME_DATA_FRAME_IN_FRAME_DURATION_TAG = "InFrameDuration";

std::string MeasurementResultsXml::MEMORY_DATA_FRAME_TAG = "MemoryData";
std::string MeasurementResultsXml::MEMORY_DATA_FRAME_START_HEAP_USAGE_TAG = "StartHeapUsage";
std::string MeasurementResultsXml::MEMORY_DATA_FRAME_END_HEAP_USAGE_TAG = "EndHeapUsage";
std::string MeasurementResultsXml::MEMORY_DATA_FRAME_HIGH_WATERMARK_TAG = "HighWatermark";
std::string MeasurementResultsXml::MEMORY_DATA_FRAME_IN_FRAME_HIGH_WATERMARK_TAG = "InFrameHighWatermark";

std::string MeasurementResultsXml::COUNTER_DATA_FRAME_TAG = "CounterData";
std::string MeasurementResultsXml::COUNTER_DATA_FRAME_COUNTERS_TAG = "Counters";
std::string MeasurementResultsXml::COUNTER_DATA_FRAME_IN_FRAME_COUNTERS_TAG = "InFrameCounters";
std::string MeasurementResultsXml::COUNTER_DATA_FRAME_COUNTER_TAG = "Counter";
std::string MeasurementResultsXml::COUNTER_DATA_FRAME_COUNTER_NAME_TAG	= "Name";
std::string MeasurementResultsXml::COUNTER_DATA_FRAME_COUNTER_VALUE_TAG = "Value";
}
