/*
 * CompactNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef COMPACT_DFA_H_
#define COMPACT_DFA_H_

#include <map>
#include <ostream>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <common/createUnique.hpp>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

#include "../../label/InitialStateLabel.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class InputAlphabet;
class States;
class FinalStates;
class InitialState;

template < class SymbolType, class StateType >
class CompactNFA final : public AutomatonBase, public alib::Components < CompactNFA < SymbolType, StateType >, SymbolType, ext::tuple < InputAlphabet >, ext::tuple < >, StateType, ext::tuple < States, FinalStates >, ext::tuple < InitialState > > {
protected:
	ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > transitions;

public:
	explicit CompactNFA ( StateType initialState );
	explicit CompactNFA ( ext::set < StateType > states, ext::set < SymbolType > inputAlphabet, StateType initialState, ext::set < StateType > finalStates );
	template < class EpsilonType >
	explicit CompactNFA ( const EpsilonNFA < SymbolType, EpsilonType, StateType > & other );
	explicit CompactNFA ( const MultiInitialStateNFA < SymbolType, StateType > & other );
	explicit CompactNFA ( const NFA < SymbolType, StateType > & other );
	explicit CompactNFA ( const DFA < SymbolType, StateType > & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const ext::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( ext::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const ext::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( ext::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const ext::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( StateType current, ext::vector < SymbolType > input, StateType next );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addTransitions ( StateType current, ext::vector < SymbolType > input, ext::set < StateType > next );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const StateType & current, const ext::vector < SymbolType > & input, const StateType & next );

	/**
	 * @return automaton transitions
	 */
	const ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > getTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > getTransitionsToState ( const StateType & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const CompactNFA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "CompactNFA";

		return xmlTagName;
	}

	static CompactNFA parse ( ext::deque < sax::Token >::iterator & input );
	static void parseTransition ( ext::deque < sax::Token >::iterator & input, CompactNFA & automaton );

	void compose ( ext::deque < sax::Token > & out ) const;
	void composeTransitions ( ext::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef CompactNFA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( CompactNFA < > ) == typeid ( CompactNFA < SymbolType, StateType > ) )
			return this;

		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		ext::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		DefaultStateType initialState = AutomatonNormalize::normalizeState ( std::move ( this->template accessElement < InitialState > ( ).get ( ) ) );
		ext::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		CompactNFA < > * res = new CompactNFA < > ( std::move ( states ), std::move ( alphabet ), std::move ( initialState ), std::move ( finalStates ) );

		for ( std::pair < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > && transition : ext::make_moveable_map ( transitions ) ) {
			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			ext::vector < DefaultSymbolType > input = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( transition.first.second ) );
			ext::set < DefaultStateType > targets = AutomatonNormalize::normalizeStates ( std::move ( transition.second ) );

			res->addTransitions ( std::move ( from ), std::move ( input ), std::move ( targets ) );
		}

		return res;
	}
};

} /* namespace automaton */

#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"

namespace automaton {

template < class SymbolType, class StateType >
CompactNFA < SymbolType, StateType >::CompactNFA ( ext::set < StateType > states, ext::set < SymbolType > inputAlphabet, StateType initialState, ext::set < StateType > finalStates ) : alib::Components < CompactNFA, SymbolType, ext::tuple < InputAlphabet >, ext::tuple < >, StateType, ext::tuple < States, FinalStates >, ext::tuple < InitialState > > ( ext::make_tuple ( std::move ( inputAlphabet ) ), ext::tuple < > ( ), ext::make_tuple ( std::move ( states ), std::move ( finalStates ) ), ext::make_tuple ( std::move ( initialState ) ) ) {
}

template < class SymbolType, class StateType >
CompactNFA < SymbolType, StateType >::CompactNFA ( StateType initialState ) : CompactNFA ( ext::set < StateType > { initialState }, ext::set < SymbolType > { }, initialState, ext::set < StateType > { } ) {
}

template < class SymbolType, class StateType >
template < class EpsilonType >
CompactNFA < SymbolType, StateType >::CompactNFA ( const EpsilonNFA < SymbolType, EpsilonType, StateType > & other ) : CompactNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		if ( transition.first.second.template is < EpsilonType > ( ) ) {
			ext::pair < StateType, ext::vector < SymbolType > > key = ext::make_pair ( transition.first.first, ext::vector < SymbolType > { } );
			transitions[key] = transition.second;
		} else {
			ext::pair < StateType, ext::vector < SymbolType > > key = ext::make_pair ( transition.first.first, ext::vector < SymbolType > { transition.first.second.template get < SymbolType > ( ) } );
			transitions[key] = transition.second;
		}
	}
}

template < class SymbolType, class StateType >
CompactNFA < SymbolType, StateType >::CompactNFA ( const MultiInitialStateNFA < SymbolType, StateType > & other ) : CompactNFA ( other.getStates ( ) + ext::set < StateType > { common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), other.getStates ( ) ) }, other.getInputAlphabet ( ), common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), other.getStates ( ) ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		ext::pair < StateType, ext::vector < SymbolType > > key = ext::make_pair ( transition.first.first, ext::vector < SymbolType > { transition.first.second } );
		transitions[key] = transition.second;
	}

	ext::pair < StateType, ext::vector < SymbolType > > key = ext::make_pair ( this->getInitialState ( ), ext::vector < SymbolType > { } );
	transitions[key] = other.getInitialStates ( );
}

template < class SymbolType, class StateType >
CompactNFA < SymbolType, StateType >::CompactNFA ( const NFA < SymbolType, StateType > & other ) : CompactNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		ext::pair < StateType, ext::vector < SymbolType > > key = ext::make_pair ( transition.first.first, ext::vector < SymbolType > { transition.first.second } );
		transitions[key] = transition.second;
	}
}

template < class SymbolType, class StateType >
CompactNFA < SymbolType, StateType >::CompactNFA ( const DFA < SymbolType, StateType > & other ) : CompactNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		ext::pair < StateType, ext::vector < SymbolType > > key = ext::make_pair ( transition.first.first, ext::vector < SymbolType > { transition.first.second } );
		transitions[key].insert ( transition.second );
	}
}

template < class SymbolType, class StateType >
AutomatonBase * CompactNFA < SymbolType, StateType >::clone ( ) const {
	return new CompactNFA ( * this );
}

template < class SymbolType, class StateType >
AutomatonBase * CompactNFA < SymbolType, StateType >::plunder ( ) && {
	return new CompactNFA ( std::move ( * this ) );
}

template < class SymbolType, class StateType >
bool CompactNFA < SymbolType, StateType >::addTransition ( StateType from, ext::vector < SymbolType > input, StateType to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist." );

	ext::set < SymbolType > inputStringAlphabet ( input.begin ( ), input.end ( ) );

	 // Transition regexp's alphabet must be subset of automaton's alphabet
	if ( !std::includes ( getInputAlphabet ( ).begin ( ), getInputAlphabet ( ).end ( ), inputStringAlphabet.begin ( ), inputStringAlphabet.end ( ) ) )
		throw AutomatonException ( "Input string is over different alphabet than automaton" );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist." );

	ext::pair < StateType, ext::vector < SymbolType > > key = ext::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

template < class SymbolType, class StateType >
void CompactNFA < SymbolType, StateType >::addTransitions ( StateType from, ext::vector < SymbolType > input, ext::set < StateType > to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist." );

	ext::set < SymbolType > inputStringAlphabet ( input.begin ( ), input.end ( ) );

	 // Transition regexp's alphabet must be subset of automaton's alphabet
	if ( !std::includes ( getInputAlphabet ( ).begin ( ), getInputAlphabet ( ).end ( ), inputStringAlphabet.begin ( ), inputStringAlphabet.end ( ) ) )
		throw AutomatonException ( "Input string is over different alphabet than automaton" );

	if ( !std::includes ( getStates ( ).begin ( ), getStates ( ).end ( ), to.begin ( ), to.end ( ) ) )
		throw AutomatonException ( "Some target states don't exist." );

	ext::pair < StateType, ext::vector < SymbolType > > key = ext::make_pair ( std::move ( from ), std::move ( input ) );

	transitions [ std::move ( key ) ].insert ( ext::make_moveable_set ( to ).begin ( ), ext::make_moveable_set ( to ).end ( ) );
}

template < class SymbolType, class StateType >
bool CompactNFA < SymbolType, StateType >::removeTransition ( const StateType & from, const ext::vector < SymbolType > & input, const StateType & to ) {
	ext::pair < StateType, ext::vector < SymbolType > > key = std::make_pair ( from, input );

	return transitions[key].erase ( to );
}

template < class SymbolType, class StateType >
const ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > & CompactNFA < SymbolType, StateType >::getTransitions ( ) const {
	return transitions;
}

template < class SymbolType, class StateType >
ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > CompactNFA < SymbolType, StateType >::getTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist" );

	ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > transitionsFromState;

	for ( const std::pair < const ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState.insert ( std::make_pair ( transition.first, transition.second ) );

	return transitionsFromState;
}

template < class SymbolType, class StateType >
ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > CompactNFA < SymbolType, StateType >::getTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist" );

	ext::map < ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > transitionsToState;

	for ( const std::pair < const ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState.insert ( std::make_pair ( transition.first, transition.second ) );

	return transitionsToState;
}

template < class SymbolType, class StateType >
int CompactNFA < SymbolType, StateType >::compare ( const CompactNFA & other ) const {
	auto first = ext::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = ext::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class StateType >
void CompactNFA < SymbolType, StateType >::operator >>( std::ostream & out ) const {
	out << "(CompactNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialState = " << getInitialState ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

template < class SymbolType, class StateType >
CompactNFA < SymbolType, StateType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class StateType >
CompactNFA < SymbolType, StateType > CompactNFA < SymbolType, StateType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, CompactNFA < SymbolType, StateType >::getXmlTagName() );

	ext::set < StateType > states = AutomatonFromXMLParser::parseStates < StateType > ( input );
	ext::set < SymbolType > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet < SymbolType > ( input );
	StateType initialState = AutomatonFromXMLParser::parseInitialState < StateType > ( input );
	ext::set < StateType > finalStates = AutomatonFromXMLParser::parseFinalStates < StateType > ( input );

	CompactNFA < SymbolType, StateType > automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < CompactNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, CompactNFA < SymbolType, StateType >::getXmlTagName() );
	return automaton;
}

template < class SymbolType, class StateType >
void CompactNFA < SymbolType, StateType >::parseTransition ( ext::deque < sax::Token >::iterator & input, CompactNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	StateType from = AutomatonFromXMLParser::parseTransitionFrom < StateType > ( input );
	ext::vector < SymbolType > inputString = AutomatonFromXMLParser::parseTransitionInputSymbolMultiple < SymbolType > ( input );
	StateType to = AutomatonFromXMLParser::parseTransitionTo < StateType > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputString ), std::move ( to ) );
}

template < class SymbolType, class StateType >
void CompactNFA < SymbolType, StateType >::compose ( ext::deque < sax::Token > & out ) const {
	out.emplace_back ( CompactNFA < SymbolType, StateType >::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( CompactNFA < SymbolType, StateType >::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class StateType >
void CompactNFA < SymbolType, StateType >::composeTransitions ( ext::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputSymbolMultiple ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class StateType >
alib::ObjectBase* CompactNFA < SymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace alib {

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::CompactNFA < SymbolType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::CompactNFA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for ( const std::pair < const ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > & transition : automaton.getTransitions ( ) ) {
			ext::set < SymbolType > alphabet ( transition.first.second.begin ( ), transition.first.second.end ( ) );
			if ( alphabet.count ( symbol ) )
				return true;
		}

		return false;
	}

	static bool available ( const automaton::CompactNFA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::CompactNFA < SymbolType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::CompactNFA < SymbolType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::CompactNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const ext::pair < StateType, ext::vector < SymbolType > >, ext::set < StateType > > & t : automaton.getTransitions ( ) )
			if ( ( t.first.first == state ) || t.second.count ( state ) )
				return true;

		return false;
	}

	static bool available ( const automaton::CompactNFA < SymbolType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::CompactNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::CompactNFA < SymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::CompactNFA < SymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::CompactNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::CompactNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class StateType >
class ElementConstraint< automaton::CompactNFA < SymbolType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::CompactNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::CompactNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

} /* namespace alib */

#endif /* COMPACT_DFA_H_ */
