/*
 * EpsilonNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef EPSILON_NFA_H_
#define EPSILON_NFA_H_

#include <map>
#include <variant>
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include <core/components.hpp>
#include <common/createUnique.hpp>
#include <object/UniqueObject.h>

#include "../AutomatonException.h"
#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

#include "../../label/InitialStateLabel.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton with epsilon transitions.
 */
class InputAlphabet;
class States;
class FinalStates;
class InitialState;

template<class SymbolType, class EpsilonType, class StateType >
class EpsilonNFA final : public AutomatonBase, public alib::Components < EpsilonNFA < SymbolType, EpsilonType, StateType >, SymbolType, ext::tuple < InputAlphabet >, ext::tuple < >, StateType, ext::tuple < States, FinalStates >, ext::tuple < InitialState > > {
protected:
	ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > transitions;

public:
	explicit EpsilonNFA ( StateType initialState );
	explicit EpsilonNFA ( ext::set < StateType > states, ext::set < SymbolType > inputAlphabet, StateType initialState, ext::set < StateType > finalStates );
	explicit EpsilonNFA ( const MultiInitialStateNFA < SymbolType, StateType > & other );
	explicit EpsilonNFA ( const NFA < SymbolType, StateType > & other );
	explicit EpsilonNFA ( const DFA < SymbolType, StateType > & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const ext::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( ext::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const ext::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( ext::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const ext::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( StateType from, ext::variant < EpsilonType, SymbolType > input, StateType to );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( StateType from, SymbolType input, StateType to );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( StateType from, StateType to );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addTransitions ( StateType from, ext::variant < EpsilonType, SymbolType > input, ext::set < StateType > to );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addTransitions ( StateType from, SymbolType input, ext::set < StateType > to );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addTransitions ( StateType from, ext::set < StateType > to );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const StateType & current, const SymbolType & input, const StateType & next );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const StateType & current, const StateType & next );

	/**
	 * @return automaton transitions
	 */
	const ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & getTransitions ( ) const;

	/**
	 * @return automaton epsilon transitions
	 */
	ext::map < StateType, ext::set < StateType > > getEpsilonTransitions ( ) const;

	/**
	 * @return automaton transitions reading some symbol
	 */
	ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > getSymbolTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > getTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton epsilon transitions from state
	 */
	ext::map < StateType, ext::set < StateType > > getEpsilonTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton transitions from state reading some symbols
	 */
	ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > getSymbolTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > getTransitionsToState ( const StateType & from ) const;

	/**
	 * @return automaton epsilon transitions to state
	 */
	ext::map < StateType, ext::set < StateType > > getEpsilonTransitionsToState ( const StateType & to ) const;

	/**
	 * @return automaton transitions to state reading some symbols
	 */
	ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > getSymbolTransitionsToState ( const StateType & to ) const;

	/**
	 * Determines whether NFA is without epsilon transitions.
	 * @return true when automaton doesn't contain epsilon transitions, false otherwise
	 */
	bool isEpsilonFree ( ) const;

	/**
	 * Determines whether NFA is deterministic.
	 * FA is deterministic if and only if:
	 * \li \c is epsilon free.
	 * \li \c size of transition function \delta (from state, input symbol) \leq 1
	 * @return true when automaton is deterministic, false otherwise
	 */
	bool isDeterministic ( ) const;

	/**
	 * Determines whether NFA is total deterministic
	 * FA is deterministic if and only if
	 * \li \c it is deterministic
	 * \li \c size of transition function \forall states and \forall input symbols \delta (from state, input symbol) = 1
	 * @return true when automaton is total deterministic, false otherwise
	 */
	bool isTotal ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const EpsilonNFA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "EpsilonNFA";

		return xmlTagName;
	}

	static EpsilonNFA parse ( ext::deque < sax::Token >::iterator & input );
	static void parseTransition ( ext::deque < sax::Token >::iterator & input, EpsilonNFA & automaton );

	void compose ( ext::deque < sax::Token > & out ) const;
	void composeTransitions ( ext::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef EpsilonNFA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( EpsilonNFA < > ) == typeid ( EpsilonNFA < SymbolType, EpsilonType, StateType > ) )
			return this;

		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		ext::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		DefaultStateType initialState = AutomatonNormalize::normalizeState ( std::move ( this->template accessElement < InitialState > ( ).get ( ) ) );
		ext::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		EpsilonNFA < > * res = new EpsilonNFA < > ( std::move ( states ), std::move ( alphabet ), std::move ( initialState ), std::move ( finalStates ) );

		for ( std::pair < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > && transition : ext::make_moveable_map ( transitions ) ) {
			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			ext::variant < DefaultEpsilonType, DefaultSymbolType > input = AutomatonNormalize::normalizeSymbolEpsilon ( std::move ( transition.first.second ) );
			ext::set < DefaultStateType > targets = AutomatonNormalize::normalizeStates ( std::move ( transition.second ) );

			res->addTransitions ( std::move ( from ), std::move ( input ), std::move ( targets ) );
		}

		return res;
	}
};

} /* namespace automaton */

#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"

namespace automaton {

template<class SymbolType, class EpsilonType, class StateType >
EpsilonNFA < SymbolType, EpsilonType, StateType >::EpsilonNFA ( ext::set < StateType > states, ext::set < SymbolType > inputAlphabet, StateType initialState, ext::set < StateType > finalStates ) : alib::Components < EpsilonNFA, SymbolType, ext::tuple < InputAlphabet >, ext::tuple < >, StateType, ext::tuple < States, FinalStates >, ext::tuple < InitialState > > ( ext::make_tuple ( std::move ( inputAlphabet ) ), ext::tuple < > ( ), ext::make_tuple ( std::move ( states ), std::move ( finalStates ) ), ext::make_tuple ( std::move ( initialState ) ) ) {
}

template<class SymbolType, class EpsilonType, class StateType >
EpsilonNFA < SymbolType, EpsilonType, StateType >::EpsilonNFA ( StateType initialState ) : EpsilonNFA ( ext::set < StateType > { initialState }, ext::set < SymbolType > { }, initialState, ext::set < StateType > { } ) {
}


template<class SymbolType, class EpsilonType, class StateType >
EpsilonNFA < SymbolType, EpsilonType, StateType >::EpsilonNFA ( const MultiInitialStateNFA < SymbolType, StateType > & other ) : EpsilonNFA ( other.getStates ( ) + ext::set < StateType > { common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), other.getStates ( ) ) }, other.getInputAlphabet ( ), common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), other.getStates ( ) ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key = ext::make_pair ( transition.first.first, ext::variant < EpsilonType, SymbolType > ( transition.first.second ) );
		transitions[key] = transition.second;
	}

	ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key = ext::make_pair ( getInitialState ( ), ext::variant < EpsilonType, SymbolType >::template from < EpsilonType > ( ) );
	transitions[key] = other.getInitialStates ( );
}


template<class SymbolType, class EpsilonType, class StateType >
EpsilonNFA < SymbolType, EpsilonType, StateType >::EpsilonNFA ( const NFA < SymbolType, StateType > & other ) : EpsilonNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key = ext::make_pair ( transition.first.first, ext::variant < EpsilonType, SymbolType > ( transition.first.second ) );
		transitions[key] = transition.second;
	}
}

template<class SymbolType, class EpsilonType, class StateType >
EpsilonNFA < SymbolType, EpsilonType, StateType >::EpsilonNFA ( const DFA < SymbolType, StateType > & other ) : EpsilonNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key = ext::make_pair ( transition.first.first, ext::variant < EpsilonType, SymbolType > ( transition.first.second ) );
		transitions[key].insert ( transition.second );
	}
}

template<class SymbolType, class EpsilonType, class StateType >
AutomatonBase * EpsilonNFA < SymbolType, EpsilonType, StateType >::clone ( ) const {
	return new EpsilonNFA ( * this );
}

template<class SymbolType, class EpsilonType, class StateType >
AutomatonBase * EpsilonNFA < SymbolType, EpsilonType, StateType >::plunder ( ) && {
	return new EpsilonNFA ( std::move ( * this ) );
}

template<class SymbolType, class EpsilonType, class StateType >
bool EpsilonNFA < SymbolType, EpsilonType, StateType >::addTransition ( StateType from, ext::variant < EpsilonType, SymbolType > input, StateType to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist." );

	if ( input.template is < SymbolType > ( ) && !getInputAlphabet ( ).count ( input.template get < SymbolType > ( ) ) )
		throw AutomatonException ( "Input symbol \"" + ext::to_string ( input.template get < SymbolType > ( ) ) + "\" doesn't exist." );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist." );

	ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key = ext::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

template<class SymbolType, class EpsilonType, class StateType >
bool EpsilonNFA < SymbolType, EpsilonType, StateType >::addTransition ( StateType from, SymbolType input, StateType to ) {
	ext::variant < EpsilonType, SymbolType > inputVariant ( std::move ( input ) );

	return addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template<class SymbolType, class EpsilonType, class StateType >
bool EpsilonNFA < SymbolType, EpsilonType, StateType >::addTransition ( StateType from, StateType to ) {
	auto inputVariant = ext::variant < EpsilonType, SymbolType >::template from < EpsilonType > ( );

	return addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template < class SymbolType, class EpsilonType, class StateType >
void EpsilonNFA < SymbolType, EpsilonType, StateType >::addTransitions ( StateType from, ext::variant < EpsilonType, SymbolType > input, ext::set < StateType > to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist." );

	if ( input.template is < SymbolType > ( ) && !getInputAlphabet ( ).count ( input.template get < SymbolType > ( ) ) )
		throw AutomatonException ( "Input symbol \"" + ext::to_string ( input.template get < SymbolType > ( ) ) + "\" doesn't exist." );

	if ( !std::includes ( getStates ( ).begin ( ), getStates ( ).end ( ), to.begin ( ), to.end ( ) ) )
		throw AutomatonException ( "Some target states don't exist." );

	ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key = ext::make_pair ( std::move ( from ), std::move ( input ) );

	transitions [ std::move ( key ) ].insert ( ext::make_moveable_set ( to ).begin ( ), ext::make_moveable_set ( to ).end ( ) );
}

template<class SymbolType, class EpsilonType, class StateType >
void EpsilonNFA < SymbolType, EpsilonType, StateType >::addTransitions ( StateType from, SymbolType input, ext::set < StateType > to ) {
	ext::variant < EpsilonType, SymbolType > inputVariant ( std::move ( input ) );

	addTransitions ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template<class SymbolType, class EpsilonType, class StateType >
void EpsilonNFA < SymbolType, EpsilonType, StateType >::addTransitions ( StateType from, ext::set < StateType > to ) {
	auto inputVariant = ext::variant < EpsilonType, SymbolType >::template from < EpsilonType > ( );

	addTransitions ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template<class SymbolType, class EpsilonType, class StateType >
bool EpsilonNFA < SymbolType, EpsilonType, StateType >::removeTransition ( const StateType & from, const SymbolType & input, const StateType & to ) {
	ext::variant < EpsilonType, SymbolType > inputVariant ( input );
	ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key = ext::make_pair ( from, inputVariant );

	return transitions[key].erase ( to );
}

template<class SymbolType, class EpsilonType, class StateType >
bool EpsilonNFA < SymbolType, EpsilonType, StateType >::removeTransition ( const StateType & from, const StateType & to ) {
	auto inputVariant = ext::variant < EpsilonType, SymbolType >::template from < EpsilonType > ( );

	ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key = ext::make_pair ( from, inputVariant );

	return transitions[key].erase ( to );
}

template<class SymbolType, class EpsilonType, class StateType >
const ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & EpsilonNFA < SymbolType, EpsilonType, StateType >::getTransitions ( ) const {
	return transitions;
}

template<class SymbolType, class EpsilonType, class StateType >
ext::map < StateType, ext::set < StateType > > EpsilonNFA < SymbolType, EpsilonType, StateType >::getEpsilonTransitions ( ) const {
	ext::map < StateType, ext::set < StateType > > result;

	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( transition.first.second.template is < EpsilonType > ( ) )
			result[transition.first.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return result;
}

template<class SymbolType, class EpsilonType, class StateType >
ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > EpsilonNFA < SymbolType, EpsilonType, StateType >::getSymbolTransitions ( ) const {
	ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > result;

	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( transition.first.second.template is < SymbolType > ( ) )
			result[ext::pair < StateType, SymbolType > ( transition.first.first, transition.first.second.template get < SymbolType > ( ) )].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return result;
}

template<class SymbolType, class EpsilonType, class StateType >
ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > EpsilonNFA < SymbolType, EpsilonType, StateType >::getTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist" );

	ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > transitionsFromState;

	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsFromState;
}

template<class SymbolType, class EpsilonType, class StateType >
ext::map < StateType, ext::set < StateType > > EpsilonNFA < SymbolType, EpsilonType, StateType >::getEpsilonTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist" );

	ext::pair < StateType, ext::variant < EpsilonType, SymbolType > > key ( from, ext::variant < EpsilonType, SymbolType >::template from < EpsilonType > ( ) );
	ext::map < StateType, ext::set < StateType > > res;
	auto transition = transitions.find ( key );

	if ( transition != transitions.end ( ) )
		res.insert ( std::make_pair ( from, transition->second ) );

	return res;
}

template<class SymbolType, class EpsilonType, class StateType >
ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > EpsilonNFA < SymbolType, EpsilonType, StateType >::getSymbolTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist" );

	ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > transitionsFromState;

	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( ( transition.first.first == from ) && transition.first.second.template is < SymbolType > ( ) )
			transitionsFromState[ext::pair < StateType, SymbolType > ( transition.first.first, transition.first.second.template get < SymbolType > ( ) )].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsFromState;
}

template<class SymbolType, class EpsilonType, class StateType >
ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > EpsilonNFA < SymbolType, EpsilonType, StateType >::getTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist" );

	ext::map < ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > transitionsToState;

	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

template<class SymbolType, class EpsilonType, class StateType >
ext::map < StateType, ext::set < StateType > > EpsilonNFA < SymbolType, EpsilonType, StateType >::getEpsilonTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist" );

	ext::map < StateType, ext::set < StateType > > transitionsToState;

	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( ( transition.second.find ( to ) != transition.second.end ( ) ) && transition.first.second.template is < EpsilonType > ( ) )
			transitionsToState[transition.first.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

template<class SymbolType, class EpsilonType, class StateType >
ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > EpsilonNFA < SymbolType, EpsilonType, StateType >::getSymbolTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist" );

	ext::map < ext::pair < StateType, SymbolType >, ext::set < StateType > > transitionsToState;

	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( ( transition.second.find ( to ) != transition.second.end ( ) ) && transition.first.second.template is < SymbolType > ( ) )
			transitionsToState[ext::pair < StateType, SymbolType > ( transition.first.first, transition.first.second.template get < SymbolType > ( ) )].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

template<class SymbolType, class EpsilonType, class StateType >
bool EpsilonNFA < SymbolType, EpsilonType, StateType >::isEpsilonFree ( ) const {
	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( transition.first.second.template is < EpsilonType > ( ) )
			return false;

	return true;
}

template<class SymbolType, class EpsilonType, class StateType >
bool EpsilonNFA < SymbolType, EpsilonType, StateType >::isDeterministic ( ) const {
	for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : transitions )
		if ( transition.second.size ( ) > 1 )
			return false;

	return isEpsilonFree ( );
}

template<class SymbolType, class EpsilonType, class StateType >
bool EpsilonNFA < SymbolType, EpsilonType, StateType >::isTotal ( ) const {
	return isDeterministic ( ) && transitions.size ( ) == getInputAlphabet ( ).size ( ) * getStates ( ).size ( );
}

template<class SymbolType, class EpsilonType, class StateType >
int EpsilonNFA < SymbolType, EpsilonType, StateType >::compare ( const EpsilonNFA & other ) const {
	auto first = ext::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = ext::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template<class SymbolType, class EpsilonType, class StateType >
void EpsilonNFA < SymbolType, EpsilonType, StateType >::operator >>( std::ostream & out ) const {
	out << "(EpsilonNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialState = " << getInitialState ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

template<class SymbolType, class EpsilonType, class StateType >
EpsilonNFA < SymbolType, EpsilonType, StateType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template<class SymbolType, class EpsilonType, class StateType >
EpsilonNFA < SymbolType, EpsilonType, StateType > EpsilonNFA < SymbolType, EpsilonType, StateType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, EpsilonNFA::getXmlTagName() );

	ext::set < StateType > states = AutomatonFromXMLParser::parseStates < StateType > ( input );
	ext::set < SymbolType > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet < SymbolType > ( input );
	StateType initialState = AutomatonFromXMLParser::parseInitialState < StateType > ( input );
	ext::set < StateType > finalStates = AutomatonFromXMLParser::parseFinalStates < StateType > ( input );

	EpsilonNFA < SymbolType, EpsilonType, StateType > automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < EpsilonNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, EpsilonNFA::getXmlTagName() );
	return automaton;
}

template<class SymbolType, class EpsilonType, class StateType >
void EpsilonNFA < SymbolType, EpsilonType, StateType >::parseTransition ( ext::deque < sax::Token >::iterator & input, EpsilonNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	StateType from = AutomatonFromXMLParser::parseTransitionFrom < StateType > ( input );
	ext::variant < EpsilonType, SymbolType > inputVariant = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol < EpsilonType, SymbolType > ( input );
	StateType to = AutomatonFromXMLParser::parseTransitionTo < StateType > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template<class SymbolType, class EpsilonType, class StateType >
void EpsilonNFA < SymbolType, EpsilonType, StateType >::compose ( ext::deque < sax::Token > & out ) const {
	out.emplace_back ( EpsilonNFA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( EpsilonNFA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template<class SymbolType, class EpsilonType, class StateType >
void EpsilonNFA < SymbolType, EpsilonType, StateType >::composeTransitions ( ext::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class EpsilonType, class StateType >
alib::ObjectBase* EpsilonNFA < SymbolType, EpsilonType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace alib {

template<class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::EpsilonNFA < SymbolType, EpsilonType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton, const SymbolType & symbol ) {
		for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second.template is < SymbolType > ( ) && ( transition.first.second.template get < SymbolType > ( ) == symbol ) )
				return true;

		return false;
	}

	static bool available ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > &, const SymbolType & ) {
	}
};

template<class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::EpsilonNFA < SymbolType, EpsilonType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const ext::pair < StateType, ext::variant < EpsilonType, SymbolType > >, ext::set < StateType > > & transition : automaton.getTransitions ( ) )
			if ( ( transition.first.first == state ) || transition.second.count ( state ) )
				return true;

		return false;
	}

	static bool available ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template<class SymbolType, class EpsilonType, class StateType >
class ComponentConstraint< automaton::EpsilonNFA < SymbolType, EpsilonType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

template<class SymbolType, class EpsilonType, class StateType >
class ElementConstraint< automaton::EpsilonNFA < SymbolType, EpsilonType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > &, const StateType & ) {
	}
};

} /* namespace alib */

#endif /* EPSILON_NFA_H_ */
