/*
 * BitParallelIndex.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef BIT_PARALLEL_INDEX_H_
#define BIT_PARALLEL_INDEX_H_

#include <string>
#include <iostream>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsMap.h>
#include <container/ObjectsVector.h>
#include <primitive/Bool.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class BitParallelIndex final : public alib::ObjectBase, public alib::Components < BitParallelIndex < SymbolType >, SymbolType, ext::tuple < GeneralAlphabet >, ext::tuple < > > {
protected:
	ext::map < SymbolType, ext::vector < bool > > m_vectors;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit BitParallelIndex ( ext::set < SymbolType > alphabet, ext::map < SymbolType, ext::vector < bool > > vectors );

	/**
	 * @return Root node of the trie
	 */
	const ext::map < SymbolType, ext::vector < bool > > & getData ( ) const;

	ext::vector < SymbolType > getString ( ) const;

	const ext::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Sets the bit vector for given symbol
	 * @param tree root node to set
	 */
	void setBitVectorForSymbol ( SymbolType symbol, ext::vector < bool > data );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const BitParallelIndex & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "BitParallelIndex";

		return xmlTagName;
	}

	static BitParallelIndex parse ( ext::deque < sax::Token >::iterator & input );

	void compose ( ext::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef BitParallelIndex < > normalized_type;

	virtual ObjectBase * normalize ( ) && {
		if ( typeid ( BitParallelIndex < > ) == typeid ( BitParallelIndex < SymbolType > ) )
			return this;

		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		ext::map < DefaultSymbolType, ext::vector < bool > > vectors;
		for ( std::pair < SymbolType, ext::vector < bool > > && vector : ext::make_moveable_map ( m_vectors ) )
			vectors.insert ( std::make_pair ( alphabet::SymbolNormalize::normalizeSymbol ( std::move ( vector.first ) ), std::move ( vector.second ) ) );

		return new BitParallelIndex < > ( std::move ( alphabet ), std::move ( vectors ) );
	}
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
BitParallelIndex < SymbolType >::BitParallelIndex ( ext::set < SymbolType > alphabet, ext::map < SymbolType, ext::vector < bool > > vectors ) : alib::Components < BitParallelIndex, SymbolType, ext::tuple < GeneralAlphabet >, ext::tuple < > > ( ext::make_tuple ( std::move ( alphabet ) ), ext::tuple < > ( ) ), m_vectors ( std::move ( vectors ) ) {
}

template < class SymbolType >
alib::ObjectBase * BitParallelIndex < SymbolType >::clone ( ) const {
	return new BitParallelIndex ( * this );
}

template < class SymbolType >
alib::ObjectBase * BitParallelIndex < SymbolType >::plunder ( ) && {
	return new BitParallelIndex ( std::move ( * this ) );
}

template < class SymbolType >
const ext::map < SymbolType, ext::vector < bool > > & BitParallelIndex < SymbolType >::getData ( ) const {
	return m_vectors;
}

template < class SymbolType >
ext::vector < SymbolType > BitParallelIndex < SymbolType >::getString ( ) const {
	ext::vector < SymbolType > res;

	unsigned index = 0;
	do {
		for ( const std::pair < const SymbolType, ext::vector < bool > > & bitVector : m_vectors )
			if ( bitVector.second.size ( ) > index && bitVector.second [ index ] ) {
				res.push_back ( bitVector.first );
				continue;
			}

	} while ( res.size ( ) == index ++ + 1 );

	return res;
}

template < class SymbolType >
void BitParallelIndex < SymbolType >::setBitVectorForSymbol ( SymbolType symbol, ext::vector < bool > data ) {
	this->m_vectors [ symbol ] = std::move ( data );
}

template < class SymbolType >
void BitParallelIndex < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(BitParallelIndex " << this->m_vectors << ")";
}

template < class SymbolType >
int BitParallelIndex < SymbolType >::compare ( const BitParallelIndex & other ) const {
	auto first = ext::tie ( getData ( ), getAlphabet ( ) );
	auto second = ext::tie ( other.getData ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
BitParallelIndex < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
BitParallelIndex < SymbolType > BitParallelIndex < SymbolType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, BitParallelIndex::getXmlTagName() );
	ext::set < SymbolType > alphabet = alib::xmlApi < ext::set < SymbolType > >::parse ( input );
	ext::map < SymbolType, ext::vector < bool > > data = alib::xmlApi < ext::map < SymbolType, ext::vector < bool > > >::parse ( input );
	BitParallelIndex < SymbolType > res ( std::move ( alphabet ), std::move ( data ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, BitParallelIndex::getXmlTagName() );
	return res;
}

template < class SymbolType >
void BitParallelIndex < SymbolType >::compose ( ext::deque < sax::Token > & out ) const {
	out.emplace_back ( BitParallelIndex::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < ext::set < SymbolType > >::compose ( out, getAlphabet ( ) );
	alib::xmlApi < ext::map < SymbolType, ext::vector < bool > > >::compose ( out, getData ( ) );
	out.emplace_back ( BitParallelIndex::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* BitParallelIndex < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace alib {

template < class SymbolType >
class ComponentConstraint < indexes::stringology::BitParallelIndex < SymbolType >, SymbolType, indexes::stringology::GeneralAlphabet > {
public:
	static bool used ( const indexes::stringology::BitParallelIndex < SymbolType > & index, const SymbolType & symbol ) {
		const ext::map < SymbolType, ext::vector < bool > > & content = index.getData ( );
		return content.find( symbol ) != content.end();
	}

	static bool available ( const indexes::stringology::BitParallelIndex < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::BitParallelIndex < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace alib */

#endif /* BIT_PARALLEL_INDEX_H_ */
