/*
 * FormalRegExp.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_REG_EXP_H_
#define FORMAL_REG_EXP_H_

#include <string>
#include <set>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include <exception/CommonException.h>

#include "../RegExpBase.h"
#include "../RegExpFeatures.h"
#include "../common/RegExpFromXMLParser.h"
#include "../common/RegExpToXMLComposer.h"

#include "FormalRegExpStructure.h"

namespace regexp {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType >
class FormalRegExp final : public RegExpBase, public alib::Components < FormalRegExp < SymbolType >, SymbolType, ext::tuple < GeneralAlphabet >, ext::tuple < > > {
protected:
	FormalRegExpStructure < SymbolType > m_regExp;

public:
	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	virtual RegExpBase * clone ( ) const;

	/**
	 * @copydoc FormalRegExpElement::plunder() const
	 */
	virtual RegExpBase * plunder ( ) &&;

	explicit FormalRegExp ( );

	explicit FormalRegExp ( ext::set < SymbolType > alphabet, FormalRegExpStructure < SymbolType > regExp );
	explicit FormalRegExp ( FormalRegExpStructure < SymbolType > regExp );
	explicit FormalRegExp ( const UnboundedRegExp < SymbolType > & other );

	/**
	 * @return Root node of the regular expression tree
	 */
	const FormalRegExpStructure < SymbolType > & getRegExp ( ) const;

	/**
	 * @return Root node of the regular expression tree
	 */
	FormalRegExpStructure < SymbolType > & getRegExp ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRegExp ( FormalRegExpStructure < SymbolType > regExp );

	virtual const ext::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const FormalRegExp & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "FormalRegExp";

		return xmlTagName;
	}

	static FormalRegExp parse ( ext::deque < sax::Token >::iterator & input );

	void compose ( ext::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef FormalRegExp < > normalized_type;

	virtual RegExpBase * normalize ( ) && {
		if ( typeid ( FormalRegExp < > ) == typeid ( FormalRegExp < SymbolType > ) )
			return this;

		return new FormalRegExp < > ( std::move ( m_regExp ).normalize ( ) );
	}
};

} /* namespace regexp */

#include "../unbounded/UnboundedRegExp.h"

namespace regexp {

template < class SymbolType >
FormalRegExp < SymbolType >::FormalRegExp ( ext::set < SymbolType > alphabet, FormalRegExpStructure < SymbolType > regExp ) : alib::Components < FormalRegExp < SymbolType >, SymbolType, ext::tuple < GeneralAlphabet >, ext::tuple < > > ( ext::make_tuple ( std::move ( alphabet ) ), ext::tuple < > ( ) ), m_regExp ( std::move ( regExp ) ) {
	if ( !this->m_regExp.getStructure ( ).checkAlphabet ( getAlphabet ( ) ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType >
FormalRegExp < SymbolType >::FormalRegExp ( ) : FormalRegExp ( ext::set < SymbolType > ( ), FormalRegExpStructure < SymbolType > ( ) ) {
}

template < class SymbolType >
FormalRegExp < SymbolType >::FormalRegExp ( FormalRegExpStructure < SymbolType > regExp ) : FormalRegExp ( regExp.getStructure ( ).computeMinimalAlphabet ( ), regExp ) {
}

template < class SymbolType >
FormalRegExp < SymbolType >::FormalRegExp ( const UnboundedRegExp < SymbolType > & other ) : FormalRegExp ( other.getAlphabet ( ), FormalRegExpStructure < SymbolType > ( other.getRegExp ( ) ) ) {
}

template < class SymbolType >
RegExpBase * FormalRegExp < SymbolType >::clone ( ) const {
	return new FormalRegExp ( * this );
}

template < class SymbolType >
RegExpBase * FormalRegExp < SymbolType >::plunder ( ) && {
	return new FormalRegExp ( std::move ( * this ) );
}

template < class SymbolType >
const FormalRegExpStructure < SymbolType > & FormalRegExp < SymbolType >::getRegExp ( ) const {
	return m_regExp;
}

template < class SymbolType >
FormalRegExpStructure < SymbolType > & FormalRegExp < SymbolType >::getRegExp ( ) {
	return m_regExp;
}

template < class SymbolType >
void FormalRegExp < SymbolType >::setRegExp ( FormalRegExpStructure < SymbolType > param ) {
	if ( !param.getStructure ( ).checkAlphabet ( getAlphabet ( ) ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );

	this->m_regExp = std::move ( param );
}

template < class SymbolType >
void FormalRegExp < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(FormalRegExp " << this->m_regExp.getStructure ( ) << ")";
}

template < class SymbolType >
int FormalRegExp < SymbolType >::compare ( const FormalRegExp & other ) const {
	auto first = ext::tie ( m_regExp.getStructure ( ), getAlphabet ( ) );
	auto second = ext::tie ( other.m_regExp.getStructure ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
FormalRegExp < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
FormalRegExp < SymbolType > FormalRegExp < SymbolType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, FormalRegExp::getXmlTagName() );

	ext::set < SymbolType > alphabet = RegExpFromXMLParser::parseAlphabet < SymbolType > ( input );

	FormalRegExpStructure < SymbolType > element ( alib::xmlApi < regexp::FormalRegExpStructure < SymbolType > >::parse ( input ) );
	FormalRegExp < SymbolType > regexp ( std::move ( alphabet ), std::move ( element ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, FormalRegExp::getXmlTagName() );

	return regexp;
}

template < class SymbolType >
void FormalRegExp < SymbolType >::compose ( ext::deque < sax::Token > & out ) const {
	out.emplace_back ( FormalRegExp::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	RegExpToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	alib::xmlApi < FormalRegExpStructure < SymbolType > >::compose ( out, this->getRegExp ( ) );
	out.emplace_back ( FormalRegExp::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* FormalRegExp < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace regexp */

namespace alib {

template < class SymbolType >
class ComponentConstraint< regexp::FormalRegExp < SymbolType >, SymbolType, regexp::GeneralAlphabet > {
public:
	static bool used ( const regexp::FormalRegExp < SymbolType > & regexp, const SymbolType & symbol ) {
		return regexp.getRegExp ( ).getStructure ( ).testSymbol ( symbol );
	}

	static bool available ( const regexp::FormalRegExp < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const regexp::FormalRegExp < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace alib */

#endif /* FORMAL_REG_EXP_H_ */
