/*
 * AdjacencyMatrixDirectedGraph.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#include "AdjacencyMatrixDirectedGraph.h"

#include <compare>

#include <object/Object.h>
#include <registration/TypeRegistration.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include "../common/GraphFromXMLParser.h"
#include "../common/GraphToXMLComposer.h"
#include "../Graph.h"
//#include "utils.h"


namespace graph {

AdjacencyMatrixDirectedGraph::AdjacencyMatrixDirectedGraph() : DirectedGraph() {
}

AdjacencyMatrixDirectedGraph::AdjacencyMatrixDirectedGraph( const DirectedGraph & graph ) {
	DirectedGraph::fromDirected( graph );
}

AdjacencyMatrixDirectedGraph::AdjacencyMatrixDirectedGraph( const UndirectedGraph & ugraph ) {
	DirectedGraph::fromUndirected( ugraph );
}

GraphBase * AdjacencyMatrixDirectedGraph::clone() const
{
	return new AdjacencyMatrixDirectedGraph( *this );
}

GraphBase * AdjacencyMatrixDirectedGraph::plunder() &&
{
	return new AdjacencyMatrixDirectedGraph( std::move( *this ) );
}

REPRESENTATION AdjacencyMatrixDirectedGraph::getRepresentation() const
{
	return REPRESENTATION::ADJACENCY_MATRIX;
}

std::string AdjacencyMatrixDirectedGraph::className() const
{
	return "AdjacencyMatrixDirectedGraph";
}

size_t AdjacencyMatrixDirectedGraph::nodeCount() const
{
	return adj.size();
}

size_t AdjacencyMatrixDirectedGraph::edgeCount() const
{
	return edges.size();
}

ext::set<Node> AdjacencyMatrixDirectedGraph::getNodes() const
{
	ext::set<Node> out;

	for (auto i : adj) {
		out.insert(i.first);
	}

	return out;
}

ext::set<DirectedEdge> AdjacencyMatrixDirectedGraph::getEdges() const
{
	return edges;
}

bool AdjacencyMatrixDirectedGraph::hasNode( const Node & node ) const
{
	return adj.find(node) != adj.end();
}

bool AdjacencyMatrixDirectedGraph::hasEdge( const DirectedEdge & edge ) const
{
	return edges.find(edge) != edges.end();
}

bool AdjacencyMatrixDirectedGraph::hasEdge( const Node & from, const Node & to ) const
{
	auto search = adj.find(from);
	if (search == adj.end()) {
		return false;
	}
	return search->second.find(to) != search->second.end();
}

ext::set<DirectedEdge> AdjacencyMatrixDirectedGraph::findEdges( const Node & from, const Node & to) const
{
	ext::set<DirectedEdge> out;

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == from && e.getToNode() == to) {
			out.insert(e);
		}
	}

	return out;
}

ext::set<Node> AdjacencyMatrixDirectedGraph::neighbors( const Node & node ) const
{
	ext::set<Node> out;

	auto search = adj.find(node);
	if (search == adj.end()) {
		return out;
	}

	for (auto i : search->second) {
		if (i.second > 0) {
			out.insert(i.first);
		}
	}

	return out;
}

ext::set<DirectedEdge> AdjacencyMatrixDirectedGraph::neighborEdges( const Node & node ) const
{
	ext::set<DirectedEdge> out;
	const ext::set<Node> &nghbrs = neighbors(node);

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == node && nghbrs.find(e.getToNode()) != nghbrs.end()) {
			out.insert(e);
		}
	}

	return out;
}

ext::set<Node> AdjacencyMatrixDirectedGraph::successors( const Node & node ) const
{
	ext::set<Node> out;

	// successors are in matrix row

	auto row = adj.find(node);
	if (row != adj.end()) {
		for (auto i : row->second) {
			if (i.second > 0) {
				out.insert(i.first);
			}
		}
	}

	return out;
}

ext::set<Node> AdjacencyMatrixDirectedGraph::predecessors( const Node & node ) const
{
	ext::set<Node> out;

	// predecessors are in matrix column

	for (auto row : adj) {
		auto i = row.second.find(node);
		if (i != row.second.end()) {
			if (i->second > 0) {
				out.insert(row.first);
			}
		}
	}

	return out;
}

ext::set<DirectedEdge> AdjacencyMatrixDirectedGraph::successorEdges( const Node & node ) const
{
	ext::set<DirectedEdge> out;
	const ext::set<Node> &nghbrs = successors(node);

	for (const DirectedEdge &e : edges) {
		if (e.getFromNode() == node && nghbrs.find(e.getToNode()) != nghbrs.end()) {
			out.insert(e);
		}
	}

	return out;
}

ext::set<DirectedEdge> AdjacencyMatrixDirectedGraph::predecessorEdges( const Node & node ) const
{
	ext::set<DirectedEdge> out;
	const ext::set<Node> &nghbrs = predecessors(node);

	for (const DirectedEdge &e : edges) {
		if (e.getToNode() == node && nghbrs.find(e.getFromNode()) != nghbrs.end()) {
			out.insert(e);
		}
	}

	return out;
}

bool AdjacencyMatrixDirectedGraph::addNode( const Node & node )
{
	if (adj.find(node) != adj.end())
		return false;

	adj[node]; // create empty map for this key
	return true;
}

bool AdjacencyMatrixDirectedGraph::addNode( const Node & node, int value )
{
	if (!addNode( node ))
		return false;

	DirectedGraph::addNode( node, value );
	return true;
}

bool AdjacencyMatrixDirectedGraph::removeNode( const Node & node )
{
	auto search = adj.find(node);
	if (search == adj.end()) {
		return false;
	}

	adj.erase(search);

	for ( auto iter = edges.begin ( ); iter != edges.end ( ); ) {
		if (iter->getFromNode() == node) {
			iter = edges.erase(iter);
		} else if (iter->getToNode() == node) {
			adj[iter->getFromNode()].erase(node);
			iter = edges.erase(iter);
		} else {
			++iter;
		}
	}

	return true;
}

bool AdjacencyMatrixDirectedGraph::addEdge( const DirectedEdge & edge )
{
	if (edges.find(edge) != edges.end())
		return false;
	if (adj.find(edge.getFromNode()) == adj.end())
		return false;
	if (adj.find(edge.getToNode()) == adj.end())
		return false;

	edges.insert(edge);

	adj[edge.getFromNode()][edge.getToNode()] += 1;
	adj[edge.getToNode()]; // insert key

	return true;
}

bool AdjacencyMatrixDirectedGraph::addEdge( const DirectedEdge & edge, int value )
{
	if (!addEdge( edge ))
		return false;

	DirectedGraph::addEdge( edge, value );
	return true;
}

bool AdjacencyMatrixDirectedGraph::removeEdge( const DirectedEdge & edge )
{
	auto search = edges.find(edge);
	if (search == edges.end()) {
		return false;
	}

	edges.erase(search);
	DirectedGraph::removeEdge(edge);

	auto &map = adj[edge.getFromNode()];
	int &val = map[edge.getToNode()];
	if (--val <= 0) {
		map.erase(edge.getToNode());
	}

	return true;
}

AdjacencyMatrixDirectedGraph AdjacencyMatrixDirectedGraph::parse( ext::deque<sax::Token>::iterator & input )
{
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, getXmlTagName());

	AdjacencyMatrixDirectedGraph graph;
	GraphFromXMLParser::parseNodes(input, graph);
	GraphFromXMLParser::parseDirectedEdges(input, graph);
	GraphFromXMLParser::parseNodeValues(input, graph);
	GraphFromXMLParser::parseDirectedEdgeValues(input, graph);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, getXmlTagName());
	return graph;
}

void AdjacencyMatrixDirectedGraph::compose( ext::deque<sax::Token> & out ) const
{
	out.emplace_back(getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	GraphToXMLComposer::composeNodes(out, this->getNodes());
	GraphToXMLComposer::composeEdges(out, this->getEdges());
	GraphToXMLComposer::composeNodeValues(out, *this);
	GraphToXMLComposer::composeEdgeValues(out, *this);

	out.emplace_back(getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

alib::ObjectBase* AdjacencyMatrixDirectedGraph::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} // namespace graph

namespace alib {

auto adjMatDirGraphType  = registration::TypeRegister < graph::Graph, graph::AdjacencyMatrixDirectedGraph > ();
auto adjMatDirGraphType2 = registration::TypeRegister < alib::Object, graph::AdjacencyMatrixDirectedGraph > ();

} /* namespace alib */
