/*
 * AllEpsilonClosure.h
 *
 *  Created on: 12. 4. 2015
 *	  Author: Jan Travnicek
 */

#ifndef EFFICIENT_ALL_EPSILON_CLOSURE_H_
#define EFFICIENT_ALL_EPSILON_CLOSURE_H_

#include <core/multipleDispatch.hpp>
#include <set>
#include <map>

#include "automaton/Automaton.h"
#include "automaton/AutomatonFeatures.h"

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <iostream>

#include <regexp/properties/RegExpEpsilon.h>

namespace automaton {

namespace properties {

namespace efficient {

class AllEpsilonClosure : public alib::SingleDispatch<AllEpsilonClosure, ext::map<DefaultStateType, ext::set<DefaultStateType>>, const automaton::AutomatonBase &> {
	template < class StateType >
	static void process(const ext::map<StateType, ext::set<StateType>>& epsilonTransitions, const StateType * state, ext::map<StateType, ext::set<StateType>>& closures, ext::set<StateType>& visited, ext::set<StateType>& nonvisited);
public:
	static ext::map<DefaultStateType, ext::set<DefaultStateType>> allEpsilonClosure( const automaton::Automaton & automaton);

	/**
	 * Computes allEpsilon closure of a state in allEpsilon nonfree automaton
	 */
	template < class SymbolType, class EpsilonType, class StateType >
	static ext::map<StateType, ext::set<StateType>> allEpsilonClosure( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & fsm);
	template < class SymbolType, class StateType >
	static ext::map<StateType, ext::set<StateType>> allEpsilonClosure( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm);
	template < class SymbolType, class StateType >
	static ext::map<StateType, ext::set<StateType>> allEpsilonClosure( const automaton::NFA < SymbolType, StateType > & fsm);
	template < class SymbolType, class StateType >
	static ext::map<StateType, ext::set<StateType>> allEpsilonClosure( const automaton::DFA < SymbolType, StateType > & fsm);
	template < class SymbolType, class StateType >
	static ext::map<StateType, ext::set<StateType>> allEpsilonClosure( const automaton::ExtendedNFA < SymbolType, StateType > & fsm);
	template < class SymbolType, class StateType >
	static ext::map<StateType, ext::set<StateType>> allEpsilonClosure( const automaton::CompactNFA < SymbolType, StateType > & fsm);
};

template < class StateType >
void AllEpsilonClosure::process(const ext::map<StateType, ext::set<StateType>>& epsilonTransitions, const StateType * state, ext::map<StateType, ext::set<StateType>>& closures, ext::set<StateType>& visited, ext::set<StateType>& nonvisited) {
	if(visited.count(*state)) return;

	state = &*visited.insert(*state).first;
	nonvisited.erase(*state);

	auto tos = epsilonTransitions.find(*state);
	if(tos == epsilonTransitions.end()) return;

	for(const StateType& to : tos->second) {
		process(epsilonTransitions, &to, closures, visited, nonvisited);
		closures[*state].insert(closures[to].begin(), closures[to].end());
	}
}

template < class SymbolType, class EpsilonType, class StateType >
ext::map<StateType, ext::set<StateType>> AllEpsilonClosure::allEpsilonClosure( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & fsm) {
	ext::map<StateType, ext::set<StateType>> res;

	for(const StateType& state : fsm.getStates())
		res[state].insert(state);

	ext::set<StateType> visited;
	ext::set<StateType> nonvisited = fsm.getStates();

	ext::map<StateType, ext::set<StateType>> epsilonTransitions = fsm.getEpsilonTransitions();

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

template < class SymbolType, class StateType >
ext::map<StateType, ext::set<StateType>> AllEpsilonClosure::allEpsilonClosure( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm) {
	ext::map<StateType, ext::set<StateType>> closure;
	for(const StateType& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

template < class SymbolType, class StateType >
ext::map<StateType, ext::set<StateType>> AllEpsilonClosure::allEpsilonClosure( const automaton::NFA < SymbolType, StateType > & fsm) {
	ext::map<StateType, ext::set<StateType>> closure;
	for(const StateType& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

template < class SymbolType, class StateType >
ext::map<StateType, ext::set<StateType>> AllEpsilonClosure::allEpsilonClosure( const automaton::DFA < SymbolType, StateType > & fsm) {
	ext::map<StateType, ext::set<StateType>> closure;
	for(const StateType& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

template < class SymbolType, class StateType >
ext::map<StateType, ext::set<StateType>> AllEpsilonClosure::allEpsilonClosure( const automaton::ExtendedNFA < SymbolType, StateType > & fsm) {
	ext::map<StateType, ext::set<StateType>> res;

	for(const StateType& state : fsm.getStates())
		res[state].insert(state);

	ext::set<StateType> visited;
	ext::set<StateType> nonvisited = fsm.getStates();

	ext::map<StateType, ext::set<StateType>> epsilonTransitions;
	for(const std::pair<const ext::pair<StateType, regexp::UnboundedRegExpStructure < SymbolType > >, ext::set<StateType> >& transition : fsm.getTransitions() )
		if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
			epsilonTransitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

template < class SymbolType, class StateType >
ext::map<StateType, ext::set<StateType>> AllEpsilonClosure::allEpsilonClosure( const automaton::CompactNFA < SymbolType, StateType > & fsm) {
	ext::map<StateType, ext::set<StateType>> res;

	for(const StateType& state : fsm.getStates())
		res[state].insert(state);

	ext::set<StateType> visited;
	ext::set<StateType> nonvisited = fsm.getStates();

	ext::map<StateType, ext::set<StateType>> epsilonTransitions;
	for(const std::pair<const ext::pair<StateType, ext::vector < SymbolType > >, ext::set<StateType> >& transition : fsm.getTransitions() )
		if( transition.first.second.size() == 0 )
			epsilonTransitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

} /* namespace efficient */

} /* namespace properties */

} /* namespace automaton */

#endif /* EFFICIENT_ALL_EPSILON_CLOSURE_H_ */
