/*
 * AutomatonFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "AutomatonFromStringParser.h"

#include "exception/CommonException.h"

#include "automaton/FSM/EpsilonNFA.h"
#include "automaton/FSM/MultiInitialStateNFA.h"
#include "automaton/FSM/NFA.h"
#include "automaton/FSM/DFA.h"
#include <automaton/Automaton.h>

#include "../StringApi.hpp"

#include <registration/AlgoRegistration.hpp>

namespace automaton {

AutomatonBase * AutomatonFromStringParser::parseAutomaton ( std::string string ) {
	std::stringstream ss ( std::move ( string ) );
	return std::move ( parseAutomaton ( ss ).getData ( ) ).plunder ( );
}

Automaton AutomatonFromStringParser::parseAutomaton(std::istream& input) {
	return parseAutomaton(input, ext::set<FEATURES>({FEATURES::EPSILON_NFA, FEATURES::MULTI_INITIAL_STATE_NFA, FEATURES::NFA, FEATURES::DFA}));
}

Automaton AutomatonFromStringParser::parseAutomaton(std::istream& input, const ext::set<FEATURES>& features) {
	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	if(token.type == AutomatonFromStringLexer::TokenType::EPSILON_NFA) {
		if(!features.count(FEATURES::EPSILON_NFA)) throw exception::CommonException("Disabled formalism EpsilonNFA");
		AutomatonFromStringLexer::putback(input, token);
		return Automaton(parseEpsilonNFA(input));
	} else if(token.type == AutomatonFromStringLexer::TokenType::MULTI_INITIAL_STATE_NFA) {
		if(!features.count(FEATURES::MULTI_INITIAL_STATE_NFA)) throw exception::CommonException("Disabled Formalism MultiInitialStateNFA");
		AutomatonFromStringLexer::putback(input, token);
		return Automaton(parseMultiInitialStateNFA(input));
	} else if(token.type == AutomatonFromStringLexer::TokenType::NFA) {
		if(!features.count(FEATURES::NFA)) throw exception::CommonException("Disabled formalism NFA");
		AutomatonFromStringLexer::putback(input, token);
		return Automaton(parseNFA(input));
	} else if(token.type == AutomatonFromStringLexer::TokenType::DFA) {
		if(!features.count(FEATURES::DFA)) throw exception::CommonException("Disabled formalism DFA");
		AutomatonFromStringLexer::putback(input, token);
		return Automaton(parseDFA(input));
	} else {
		throw exception::CommonException("Formalism not recognised (token = \"" + token.value + "\")");
	}
}

EpsilonNFA < > AutomatonFromStringParser::parseEpsilonNFA(std::istream& input) {
	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	if(token.type != AutomatonFromStringLexer::TokenType::EPSILON_NFA) {
		throw exception::CommonException("Unrecognised ENFA token.");
	}
	ext::vector<ext::variant<DefaultEpsilonType, DefaultSymbolType> > symbols;

	token = AutomatonFromStringLexer::next(input);
	while(token.type != AutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(token.type == AutomatonFromStringLexer::TokenType::ERROR) {
			AutomatonFromStringLexer::putback(input, token);
			DefaultSymbolType symbol = alib::stringApi<DefaultSymbolType>::parse(input);

			ext::variant<DefaultEpsilonType, DefaultSymbolType> symbolVariant(symbol);
			symbols.push_back(symbolVariant);
		} else if(token.type == AutomatonFromStringLexer::TokenType::EPSILON) {
			symbols.push_back(ext::variant<DefaultEpsilonType, DefaultSymbolType>::from < DefaultEpsilonType > ( ));
		}

		token = AutomatonFromStringLexer::next(input);
	}

	DefaultStateType* initialState = NULL;
	ext::set<DefaultStateType> finalStates;
	ext::set<DefaultStateType> states;
	ext::set<ext::tuple<DefaultStateType, ext::variant<DefaultEpsilonType, DefaultSymbolType>, DefaultStateType>> transitionFunction;

	parseEpsilonNFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
	token = AutomatonFromStringLexer::next(input);

	while(token.type == AutomatonFromStringLexer::TokenType::NEW_LINE) {
		token = AutomatonFromStringLexer::next(input);
		if(token.type == AutomatonFromStringLexer::TokenType::TEOF)
			break;
		else
			AutomatonFromStringLexer::putback(input, token);

		parseEpsilonNFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
		token = AutomatonFromStringLexer::next(input);
	}

	if(initialState == NULL) throw exception::CommonException("No initial state recognised.");

	EpsilonNFA < > res(*initialState);
	delete initialState;

	for(const ext::variant<DefaultEpsilonType, DefaultSymbolType> inputSymbol : symbols) {
		if(inputSymbol.is<DefaultSymbolType>())
			res.addInputSymbol(inputSymbol.get<DefaultSymbolType>());
	}
	res.setStates(states);
	res.setFinalStates(finalStates);
	for(const ext::tuple<DefaultStateType, ext::variant<DefaultEpsilonType, DefaultSymbolType>, DefaultStateType> transition : transitionFunction) {
		res.addTransition(std::get<0>(transition), std::get<1>(transition), std::get<2>(transition));
	}

	return res;
}

MultiInitialStateNFA < > AutomatonFromStringParser::parseMultiInitialStateNFA(std::istream& input) {
	MultiInitialStateNFA < > res;

	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	if(token.type != AutomatonFromStringLexer::TokenType::MULTI_INITIAL_STATE_NFA) {
		throw exception::CommonException("Unrecognised MISNFA token.");
	}
	ext::vector<DefaultSymbolType> symbols;

	token = AutomatonFromStringLexer::next(input);
	while(token.type != AutomatonFromStringLexer::TokenType::NEW_LINE) {
		AutomatonFromStringLexer::putback(input, token);
		DefaultSymbolType symbol = alib::stringApi<DefaultSymbolType>::parse(input);
		res.addInputSymbol(symbol);
		symbols.push_back(symbol);

		token = AutomatonFromStringLexer::next(input);
	}

	parseMultiInitialStateNFATransition(input, res, symbols);
	token = AutomatonFromStringLexer::next(input);

	while(token.type == AutomatonFromStringLexer::TokenType::NEW_LINE) {
		token = AutomatonFromStringLexer::next(input);
		if(token.type == AutomatonFromStringLexer::TokenType::TEOF)
			break;
		else
			AutomatonFromStringLexer::putback(input, token);

		parseMultiInitialStateNFATransition(input, res, symbols);
		token = AutomatonFromStringLexer::next(input);
	}

	return res;
}

NFA < > AutomatonFromStringParser::parseNFA(std::istream& input) {
	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	if(token.type != AutomatonFromStringLexer::TokenType::NFA) {
		throw exception::CommonException("Unrecognised NFA token.");
	}
	ext::vector<DefaultSymbolType> symbols;

	token = AutomatonFromStringLexer::next(input);
	while(token.type != AutomatonFromStringLexer::TokenType::NEW_LINE) {
		AutomatonFromStringLexer::putback(input, token);
		DefaultSymbolType symbol = alib::stringApi<DefaultSymbolType>::parse(input);
		symbols.push_back(symbol);

		token = AutomatonFromStringLexer::next(input);
	}

	DefaultStateType* initialState = NULL;
	ext::set<DefaultStateType> finalStates;
	ext::set<DefaultStateType> states;
	ext::set<ext::tuple<DefaultStateType, DefaultSymbolType, DefaultStateType>> transitionFunction;

	parseNFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
	token = AutomatonFromStringLexer::next(input);

	while(token.type == AutomatonFromStringLexer::TokenType::NEW_LINE) {
		token = AutomatonFromStringLexer::next(input);
		if(token.type == AutomatonFromStringLexer::TokenType::TEOF)
			break;
		else
			AutomatonFromStringLexer::putback(input, token);

		parseNFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
		token = AutomatonFromStringLexer::next(input);
	}

	if(initialState == NULL) throw exception::CommonException("No initial state recognised.");

	NFA < > res(*initialState);
	delete initialState;

	res.setInputAlphabet(ext::set<DefaultSymbolType>(symbols.begin(), symbols.end()));
	res.setStates(states);
	res.setFinalStates(finalStates);
	for(const ext::tuple<DefaultStateType, DefaultSymbolType, DefaultStateType> transition : transitionFunction) {
		res.addTransition(std::get<0>(transition), std::get<1>(transition), std::get<2>(transition));
	}

	return res;
}

DFA < > AutomatonFromStringParser::parseDFA(std::istream& input) {
	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	if(token.type != AutomatonFromStringLexer::TokenType::DFA) {
		throw exception::CommonException("Unrecognised DFA token.");
	}
	ext::vector<DefaultSymbolType> symbols;

	token = AutomatonFromStringLexer::next(input);
	while(token.type != AutomatonFromStringLexer::TokenType::NEW_LINE) {
		AutomatonFromStringLexer::putback(input, token);
		DefaultSymbolType symbol = alib::stringApi<DefaultSymbolType>::parse(input);
		symbols.push_back(symbol);

		token = AutomatonFromStringLexer::next(input);
	}

	DefaultStateType* initialState = NULL;
	ext::set<DefaultStateType> finalStates;
	ext::set<DefaultStateType> states;
	ext::set<ext::tuple<DefaultStateType, DefaultSymbolType, DefaultStateType>> transitionFunction;

	parseDFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
	token = AutomatonFromStringLexer::next(input);

	while(token.type == AutomatonFromStringLexer::TokenType::NEW_LINE) {
		token = AutomatonFromStringLexer::next(input);
		if(token.type == AutomatonFromStringLexer::TokenType::TEOF)
			break;
		else
			AutomatonFromStringLexer::putback(input, token);

		parseDFATransition(input, states, symbols, initialState, finalStates, transitionFunction);
		token = AutomatonFromStringLexer::next(input);
	}

	if(initialState == NULL) throw exception::CommonException("No initial state recognised.");

	DFA < > res(*initialState);
	delete initialState;

	res.setInputAlphabet(ext::set<DefaultSymbolType>(symbols.begin(), symbols.end()));
	res.setStates(states);
	res.setFinalStates(finalStates);
	for(const ext::tuple<DefaultStateType, DefaultSymbolType, DefaultStateType> transition : transitionFunction) {
		res.addTransition(std::get<0>(transition), std::get<1>(transition), std::get<2>(transition));
	}

	return res;
}

void AutomatonFromStringParser::initialFinalState(std::istream& input, bool& initial, bool& final) {
	initial = false;
	final = false;

	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	if(token.type == AutomatonFromStringLexer::TokenType::IN) {
		initial = true;
		token = AutomatonFromStringLexer::next(input);

		if(token.type == AutomatonFromStringLexer::TokenType::OUT) {
			final = true;
		} else {
			AutomatonFromStringLexer::putback(input, token);
		}
	} else if(token.type == AutomatonFromStringLexer::TokenType::OUT) {
		final = true;
		token = AutomatonFromStringLexer::next(input);

		if(token.type == AutomatonFromStringLexer::TokenType::IN) {
			initial = true;
		} else {
			AutomatonFromStringLexer::putback(input, token);
		}
	} else {
		AutomatonFromStringLexer::putback(input, token);
	}
}


void AutomatonFromStringParser::parseEpsilonNFATransition(std::istream& input, ext::set<DefaultStateType>& states, const ext::vector<ext::variant<DefaultEpsilonType, DefaultSymbolType>>& symbols, DefaultStateType*& initialState, ext::set<DefaultStateType>& finalStates, ext::set<ext::tuple<DefaultStateType, ext::variant<DefaultEpsilonType, DefaultSymbolType>, DefaultStateType>>& transitionFunction) {
	bool initial = false;
	bool final = false;

	initialFinalState(input, initial, final);

	DefaultStateType from(alib::stringApi<DefaultStateType>::parse(input));
	states.insert(from);
	if(initial) {
		if(initialState != NULL) throw exception::CommonException("Multiple initial states are not avaiable for NFA type");
		initialState = new DefaultStateType(from);
	}
	if(final) finalStates.insert(from);

	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	ext::vector<ext::variant<DefaultEpsilonType, DefaultSymbolType>>::const_iterator iter = symbols.begin();

	while(token.type != AutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(iter == symbols.end()) throw exception::CommonException("Invalid line format");

		if(token.type != AutomatonFromStringLexer::TokenType::NONE) {
			AutomatonFromStringLexer::putback(input, token);
			do {
				DefaultStateType to(alib::stringApi<DefaultStateType>::parse(input));
				states.insert(to);
				transitionFunction.insert(ext::make_tuple(from, *iter, to));

				token = AutomatonFromStringLexer::next(input);
				if(token.type != AutomatonFromStringLexer::TokenType::SEPARATOR) break;
			} while(true);
		} else {
			token = AutomatonFromStringLexer::next(input);
		}
		++iter;
	}
	AutomatonFromStringLexer::putback(input, token);

	if(iter != symbols.end()) throw exception::CommonException("Invalid line format");
}

void AutomatonFromStringParser::parseMultiInitialStateNFATransition(std::istream& input, MultiInitialStateNFA <DefaultSymbolType, DefaultStateType > & res, const ext::vector<DefaultSymbolType>& symbols) {
	bool initial = false;
	bool final = false;

	initialFinalState(input, initial, final);

	DefaultStateType from(alib::stringApi<DefaultStateType>::parse(input));
	res.addState(from);
	if(initial) res.addInitialState(from);
	if(final) res.addFinalState(from);

	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	ext::vector<DefaultSymbolType>::const_iterator iter = symbols.begin();

	while(token.type != AutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(iter == symbols.end()) throw exception::CommonException("Invalid line format");

		if(token.type != AutomatonFromStringLexer::TokenType::NONE) {
			AutomatonFromStringLexer::putback(input, token);
			do {
				DefaultStateType to(alib::stringApi<DefaultStateType>::parse(input));
				res.addState(to);
				res.addTransition(from, *iter, to);

				token = AutomatonFromStringLexer::next(input);
				if(token.type != AutomatonFromStringLexer::TokenType::SEPARATOR) break;
			} while(true);
		} else {
			token = AutomatonFromStringLexer::next(input);
		}
		++iter;
	}
	AutomatonFromStringLexer::putback(input, token);

	if(iter != symbols.end()) throw exception::CommonException("Invalid line format");
}

void AutomatonFromStringParser::parseNFATransition(std::istream& input, ext::set<DefaultStateType>& states, const ext::vector<DefaultSymbolType>& symbols, DefaultStateType*& initialState, ext::set<DefaultStateType>& finalStates, ext::set<ext::tuple<DefaultStateType, DefaultSymbolType, DefaultStateType>>& transitionFunction) {
	bool initial = false;
	bool final = false;

	initialFinalState(input, initial, final);

	DefaultStateType from(alib::stringApi<DefaultStateType>::parse(input));
	states.insert(from);
	if(initial) {
		if(initialState != NULL) throw exception::CommonException("Multiple initial states are not avaiable for NFA type");
		initialState = new DefaultStateType(from);
	}
	if(final) finalStates.insert(from);

	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	ext::vector<DefaultSymbolType>::const_iterator iter = symbols.begin();

	while(token.type != AutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(iter == symbols.end()) throw exception::CommonException("Invalid line format");

		if(token.type != AutomatonFromStringLexer::TokenType::NONE) {
			AutomatonFromStringLexer::putback(input, token);
			do {
				DefaultStateType to(alib::stringApi<DefaultStateType>::parse(input));
				states.insert(to);
				transitionFunction.insert(ext::make_tuple(from, *iter, to));

				token = AutomatonFromStringLexer::next(input);
				if(token.type != AutomatonFromStringLexer::TokenType::SEPARATOR) break;
			} while(true);
		} else {
			token = AutomatonFromStringLexer::next(input);
		}
		++iter;
	}
	AutomatonFromStringLexer::putback(input, token);

	if(iter != symbols.end()) throw exception::CommonException("Invalid line format");
}

void AutomatonFromStringParser::parseDFATransition(std::istream& input, ext::set<DefaultStateType>& states, const ext::vector<DefaultSymbolType>& symbols, DefaultStateType*& initialState, ext::set<DefaultStateType>& finalStates, ext::set<ext::tuple<DefaultStateType, DefaultSymbolType, DefaultStateType>>& transitionFunction) {
	bool initial = false;
	bool final = false;

	initialFinalState(input, initial, final);

	DefaultStateType from(alib::stringApi<DefaultStateType>::parse(input));
	states.insert(from);
	if(initial) {
		if(initialState != NULL) throw exception::CommonException("Multiple initial states are not avaiable for NFA type");
		initialState = new DefaultStateType(from);
	}
	if(final) finalStates.insert(from);

	AutomatonFromStringLexer::Token token = AutomatonFromStringLexer::next(input);
	ext::vector<DefaultSymbolType>::const_iterator iter = symbols.begin();

	while(token.type != AutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(iter == symbols.end()) throw exception::CommonException("Invalid line format");

		if(token.type != AutomatonFromStringLexer::TokenType::NONE) {
			AutomatonFromStringLexer::putback(input, token);
			DefaultStateType to(alib::stringApi<DefaultStateType>::parse(input));
			states.insert(to);
			transitionFunction.insert(ext::make_tuple(from, *iter, to));

			token = AutomatonFromStringLexer::next(input);
		} else {
			token = AutomatonFromStringLexer::next(input);
		}
		++iter;
	}
	AutomatonFromStringLexer::putback(input, token);

	if(iter != symbols.end()) throw exception::CommonException("Invalid line format");
}

} /* namespace automaton */

auto AutomatonFromStringParser = registration::AbstractRegister < automaton::AutomatonFromStringParser, automaton::AutomatonBase *, std::string > ( automaton::AutomatonFromStringParser::parseAutomaton );
