/*
 * RegExpFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "RegExpFromStringParser.h"
#include "exception/CommonException.h"
#include "regexp/RegExpClasses.h"

#include "regexp/unbounded/UnboundedRegExpElements.h"
#include "regexp/formal/FormalRegExpElements.h"
#include <regexp/RegExp.h>

#include "../StringApi.hpp"

namespace regexp {

RegExp RegExpFromStringParser::parseRegExp(std::istream& input) {
	return parseRegExp(input, ext::set<FEATURES>({FEATURES::FORMAL, FEATURES::UNBOUNDED}));
}

UnboundedRegExpStructure < DefaultSymbolType > RegExpFromStringParser::parseUnboundedRegExpStructure(std::istream& input) {
	ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > element = alternation(input);
	return regexp::UnboundedRegExpStructure < DefaultSymbolType >(std::move(element));
}

RegExp RegExpFromStringParser::parseRegExp(std::istream& input, const ext::set<FEATURES>& features) {
	UnboundedRegExp < > regexp { RegExpFromStringParser::parseUnboundedRegExpStructure ( input ) };

	if(features.count(FEATURES::UNBOUNDED)) return RegExp { std::move ( regexp ) };

	if(features.count(FEATURES::FORMAL)) return RegExp { FormalRegExp < > { std::move ( regexp ) } };

	throw exception::CommonException("Invalid input");
}

ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > RegExpFromStringParser::alternation(std::istream& input) {
	return alternationCont(input, concatenation(input));
}

ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > RegExpFromStringParser::alternationCont(std::istream& input, ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > left) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::PLUS) {
		UnboundedRegExpAlternation < DefaultSymbolType > res;
		res.appendElement(std::move(left));
		res.appendElement(concatenation(input));

		return alternationContCont(input, std::move ( res ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return left;
	}
}

ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > RegExpFromStringParser::alternationContCont(std::istream& input, UnboundedRegExpAlternation < DefaultSymbolType > res) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::PLUS) {
		res.appendElement(concatenation(input));

		return alternationContCont(input, std::move ( res ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > ( std::move ( res ).plunder() );
	}
}

ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > RegExpFromStringParser::concatenation(std::istream& input) {
	return concatenationCont(input, factor(input));
}

ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > RegExpFromStringParser::concatenationCont(std::istream& input, ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > left) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::ERROR || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::EMPTY) {
		RegExpFromStringLexer::putback(input, token);
		UnboundedRegExpConcatenation < DefaultSymbolType > res;
		res.appendElement(std::move(left));
		res.appendElement(factor(input));

		return concatenationContCont(input, std::move ( res ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return left;
	}
}

ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > RegExpFromStringParser::concatenationContCont(std::istream& input, UnboundedRegExpConcatenation < DefaultSymbolType > res) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::ERROR || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::EMPTY) {
		RegExpFromStringLexer::putback(input, token);
		res.appendElement(factor(input));

		return concatenationContCont(input, std::move ( res ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > ( std::move ( res ).plunder() );
	}
}

ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > RegExpFromStringParser::factor(std::istream& input) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::LPAR) {
		ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > base = alternation(input);
		token = RegExpFromStringLexer::next(input);
		if(token.type != RegExpFromStringLexer::TokenType::RPAR) throw exception::CommonException("Expected RPAR");
		return star(input, std::move ( base ) );
	} else if(token.type == RegExpFromStringLexer::TokenType::EPS) {
		return star(input, ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > ( new UnboundedRegExpEpsilon < DefaultSymbolType > ( ) ) );
	} else if(token.type == RegExpFromStringLexer::TokenType::EMPTY) {
		return star(input, ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > ( new UnboundedRegExpEmpty < DefaultSymbolType > ( ) ) );
	} else if(token.type == RegExpFromStringLexer::TokenType::ERROR) {
		UnboundedRegExpSymbol < DefaultSymbolType > res(alib::stringApi<DefaultSymbolType>::parse(input));
		return star(input, std::move ( res ) );
	} else {
		throw exception::CommonException("Unrecognised token at factor rule");
	}
}

ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > RegExpFromStringParser::star(std::istream& input, UnboundedRegExpElement < DefaultSymbolType > && elem) {
	RegExpFromStringLexer::Token token = RegExpFromStringLexer::next(input);
	if(token.type == RegExpFromStringLexer::TokenType::STAR) {
		UnboundedRegExpIteration < DefaultSymbolType > iter ( std::move ( elem ) );
		return star(input, std::move ( iter ) );
	} else {
		RegExpFromStringLexer::putback(input, token);
		return ext::rvalue_ref < UnboundedRegExpElement < DefaultSymbolType > > ( std::move ( elem ).plunder() );
	}
}

}
