/*
 * atrim.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Tomas Pecka
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>
#include <sax/FromXMLParserHelper.h>

#include <factory/XmlDataFactory.hpp>
#include <exception/CommonException.h>

#include "grammar/simplify/Trim.h"
#include "grammar/simplify/UnproductiveSymbolsRemover.h"
#include "grammar/simplify/UnreachableSymbolsRemover.h"
#include "automaton/simplify/Trim.h"
#include "automaton/simplify/efficient/Trim.h"
#include "automaton/simplify/UselessStatesRemover.h"
#include "automaton/simplify/efficient/UselessStatesRemover.h"
#include "automaton/simplify/UnreachableStatesRemover.h"
#include "automaton/simplify/efficient/UnreachableStatesRemover.h"
#include "regexp/simplify/RegExpOptimize.h"

grammar::Grammar trimGrammar(const grammar::Grammar& g, bool del_unreachable, bool del_unproductive) {
	if( del_unreachable && del_unproductive )
		return grammar::simplify::Trim::trim( g );
	if( del_unreachable )
		return grammar::simplify::UnreachableSymbolsRemover::remove( g );
	if( del_unproductive )
		return grammar::simplify::UnproductiveSymbolsRemover::remove( g );
	return grammar::simplify::Trim::trim( g );
}

automaton::Automaton trimAutomaton(const automaton::Automaton& g, bool del_unreachable, bool del_useless) {
	if( del_unreachable && del_useless )
		return automaton::simplify::Trim::trim( g );
	if( del_unreachable )
		return automaton::simplify::UnreachableStatesRemover::remove( g );
	if( del_useless )
		return automaton::simplify::UselessStatesRemover::remove( g );
	return automaton::simplify::Trim::trim( g );
}

automaton::Automaton efficientTrimAutomaton(const automaton::Automaton& g, bool del_unreachable, bool del_useless) {
	if( del_unreachable && del_useless )
		return automaton::simplify::efficient::Trim::trim( g );
	if( del_unreachable )
		return automaton::simplify::efficient::UnreachableStatesRemover::remove( g );
	if( del_useless )
		return automaton::simplify::efficient::UselessStatesRemover::remove( g );
	return automaton::simplify::efficient::Trim::trim( g );
}

regexp::RegExp optimizeRegExp(const regexp::RegExp& r) {
	return regexp::simplify::RegExpOptimize::optimize( r );
}

int main(int argc, char* argv[]) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd("Removes unreachable and useless states from FSM, productive and unreachable nonterminals from CFG. Simplifies representation of RE", ' ', "0.01");

		TCLAP::SwitchArg useless(		"u",	"useless",	"Removes useless states (or symbols). (works with FSM, FTA or CFG)"	);
		cmd.add( useless );

		TCLAP::SwitchArg unreachable(		"r",	"unreachable",	"Removes unreachable states (or symbols). (works with FSM, FTA or CFG)"	);
		cmd.add( unreachable );

		TCLAP::SwitchArg simplify(		"s",	"simplify",	"Simplifies representation. (works with RE only)"			);
		cmd.add( simplify );

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Input to trim",							false,	"-",	"file");
		cmd.add( input );

		TCLAP::SwitchArg efficient(		"e",	"efficient",	"Use efficient implementation",		false);
		cmd.add( efficient );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		measurements::start("Overal", measurements::Type::OVERALL);
		measurements::start("Input read", measurements::Type::AUXILIARY);

		ext::deque<sax::Token> tokens = sax::FromXMLParserHelper::parseInput(input);

		if( alib::XmlDataFactory::first<automaton::Automaton>(tokens)) {
			automaton::Automaton automaton = alib::XmlDataFactory::fromTokens (std::move(tokens));

			measurements::end();
			measurements::start("Algorithm", measurements::Type::MAIN);

			if(efficient.getValue()) {
				automaton::Automaton res = efficientTrimAutomaton(automaton, unreachable.getValue(), useless.getValue() );

				measurements::end();
				measurements::start("Output write", measurements::Type::AUXILIARY);

				alib::XmlDataFactory::toStdout( res );
			} else {
				automaton::Automaton res = trimAutomaton(automaton, unreachable.getValue(), useless.getValue() );

				measurements::end();
				measurements::start("Output write", measurements::Type::AUXILIARY);

				alib::XmlDataFactory::toStdout( res );
			}
		} else if( alib::XmlDataFactory::first<grammar::Grammar>(tokens)) {
			grammar::Grammar grammar = alib::XmlDataFactory::fromTokens (std::move(tokens));

			measurements::end();
			measurements::start("Algorithm", measurements::Type::MAIN);

			grammar::Grammar res = trimGrammar(grammar, unreachable.getValue(), useless.getValue() );

			measurements::end();
			measurements::start("Output write", measurements::Type::AUXILIARY);

			alib::XmlDataFactory::toStdout( res );
		} else if( alib::XmlDataFactory::first<regexp::RegExp>(tokens)) {
			regexp::RegExp regexp = alib::XmlDataFactory::fromTokens (std::move(tokens));

			measurements::end();
			measurements::start("Algorithm", measurements::Type::MAIN);

			regexp::RegExp res = optimizeRegExp( regexp );

			measurements::end();
			measurements::start("Output write", measurements::Type::AUXILIARY);

			alib::XmlDataFactory::toStdout( res );
		} else {
			throw exception::CommonException( "Input not recognized." );
		}

		measurements::end();
		measurements::end();

		if(measure.getValue()) ext::cmeasure << measurements::results() << std::endl;

		return 0;
	} catch( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
